#!/usr/bin/env python

#******************************************************************************
#
#  rpnLexicographic.py
#
#  rpnChilada lexicography operators
#  copyright (c) 2020, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

import itertools
import string

from mpmath import arange, fadd, ceil, floor, fmod, fmul, fneg, fprod, fsub, \
                   fsum, log10, mp, mpf, mpmathify, nint, power

import gmpy2

from rpn.rpnBase import convertToBaseN
from rpn.rpnFactor import getFactors
from rpn.rpnGenerator import RPNGenerator
from rpn.rpnMath import isDivisible
from rpn.rpnNumberTheory import getPowMod
from rpn.rpnPersistence import cachedFunction
from rpn.rpnPrimeUtils import isPrime
from rpn.rpnUtils import oneArgFunctionEvaluator, twoArgFunctionEvaluator, validateRealInt, \
                         getMPFIntegerAsString


#******************************************************************************
#
#  splitNumberByDigits
#
#  This splits the number into 2 numbers by splitting off half the digits
#  for the first and the rest of the digits for the second.  If there is an
#  odd number of digits then the extra digit goes to the second number.
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def splitNumberByDigits( n ):
    n = getMPFIntegerAsString( n )

    split = len( n ) // 2

    return mpmathify( n[ : split ] ), mpmathify( n[ split : ] )


#******************************************************************************
#
#  getDigitList
#
#******************************************************************************

def getDigitList( n, dropZeroes = False ):
    result = [ ]

    for c in getMPFIntegerAsString( n ):
        if dropZeroes and c == '0':
            continue

        result.append( int( c ) )

    return result


@oneArgFunctionEvaluator( )
def getDigits( n ):
    return getDigitList( n )

@oneArgFunctionEvaluator( )
def getNonzeroDigits( n ):
    return getDigitList( n, dropZeroes = True )


#******************************************************************************
#
#  getRightDigits
#
#******************************************************************************

@twoArgFunctionEvaluator( )
def getRightDigits( n, k ):
    return fmod( n, pow( 10, k ) )


#******************************************************************************
#
#  getLeftDigits
#
#******************************************************************************

@twoArgFunctionEvaluator( )
def getLeftDigits( n, k ):
    return combineDigits( getDigitList( n )[ : int( k ) ] )


#******************************************************************************
#
#  getBaseKDigitList
#
#  gmpy2's digits( ) function is way faster, but only works for bases 2 to
#  62.  I need to overhaul convertToBaseN( )... it's way too slow.
#
#******************************************************************************

def getBaseKDigitList( n, base, dropZeroes = False ):
    if 1 < base < 63:
        if n < 0:
            raise ValueError( '\'get_base_k_digits\' does not support negative numbers.' )

        asciiDigits = gmpy2.digits( int( n ), int( base ) )

        digits = [ ]

        ord0 = ord( '0' )
        orda = ord( 'a' )
        ordA = ord( 'A' )

        for i in asciiDigits:
            if '0' <= i <= '9':
                digits.append( ord( i ) - ord0 )
            elif 'a' <= i <= 'z':
                digits.append( ord( i ) - orda + 10 )
            else:
                digits.append( ord( i ) - ordA + 36 )
    else:
        digits = convertToBaseN( n, base, outputBaseDigits=True )

    result = [ ]

    for digit in digits:
        if dropZeroes and digit == 0:
            continue

        result.append( digit )

    return result


@twoArgFunctionEvaluator( )
def getBaseKDigits( n, k ):
    return getBaseKDigitList( n, k )

@twoArgFunctionEvaluator( )
def getNonzeroBaseKDigits( n, k ):
    return getBaseKDigitList( n, k, dropZeroes = True )


#******************************************************************************
#
#  isBaseKPandigital
#
#******************************************************************************

@twoArgFunctionEvaluator( )
def isBaseKPandigital( n, base ):
    digits = convertToBaseN( n, base, outputBaseDigits=True )

    for i in arange( min( int( base ), len( digits ) ) ):
        try:
            digits.index( i )
        except:
            return 0

    return 1


#******************************************************************************
#
#  sumDigits
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def sumDigits( n ):
    result = 0

    for c in getMPFIntegerAsString( n ):
        result = fadd( result, int( c ) )

    return result


#******************************************************************************
#
#  multiplyDigitList
#
#******************************************************************************

def multiplyDigitList( n, exponent = 1, dropZeroes = False ):
    if exponent < 1:
        raise ValueError( 'multiplyDigitList( ) expects a positive integer for \'exponent\'' )

    if exponent == 1:
        return fprod( getDigitList( n, dropZeroes ) )

    return fprod( [ power( i, exponent ) for i in getDigitList( n, dropZeroes ) ] )

@oneArgFunctionEvaluator( )
def multiplyDigits( n ):
    return multiplyDigitList( n )

@twoArgFunctionEvaluator( )
def multiplyDigitPowers( n, k ):
    return multiplyDigitList( n, k )

@oneArgFunctionEvaluator( )
def multiplyNonzeroDigits( n ):
    return multiplyDigitList( n, dropZeroes = True )

@twoArgFunctionEvaluator( )
def multiplyNonzeroDigitPowers( n, k ):
    return multiplyDigitList( n, k, dropZeroes = True )


#******************************************************************************
#
#  appendDigits
#
#******************************************************************************

def appendDigits( n, digits, digitCount ):
    if validateRealInt( n ) < 0:
        return nint( fsub( fmul( floor( n ), power( 10, digitCount ) ), digits ) )
    else:
        return nint( fadd( fmul( floor( n ), power( 10, digitCount ) ), digits ) )


#******************************************************************************
#
#  addDigits
#
#******************************************************************************

@twoArgFunctionEvaluator( )
def addDigits( n, k ):
    if validateRealInt( k ) < 0:
        raise ValueError( "'add_digits' requires a non-negative integer for the second argument" )

    digits = int( k )

    if digits == 0:
        digitCount = 1
    else:
        digitCount = int( fadd( floor( log10( k ) ), 1 ) )

    return appendDigits( n, digits, digitCount )


#******************************************************************************
#
#  combineDigits
#
#******************************************************************************

def combineDigits( n ):
    if isinstance( n, RPNGenerator ):
        return combineDigits( list( n ) )

    result = 0

    listResult = False

    for i in n:
        if isinstance( i, ( list, RPNGenerator ) ) and result == 0:
            listResult = True
            result = [ combineDigits( i ) ]
        elif listResult:
            if not isinstance( result, list ):
                result = [ result ]

            result.append( combineDigits( i ) )
        else:
            result = addDigits( result, validateRealInt( i ) )

    return result


#******************************************************************************
#
#  replaceDigits
#
#******************************************************************************

def replaceDigits( n, source, replace ):
    n = getMPFIntegerAsString( validateRealInt( n ) )

    if  validateRealInt( source ) < 0:
        raise ValueError( 'source value must be a positive integer' )

    if validateRealInt( replace ) < 0:
        raise ValueError( 'replace value must be a positive integer' )

    return mpmathify( n.replace( str( int ( source ) ), str( int( replace ) ) ) )


#******************************************************************************
#
#  duplicateDigits
#
#******************************************************************************

@twoArgFunctionEvaluator( )
def duplicateDigits( n, k ):
    if validateRealInt( k ) < 0:
        raise ValueError( "'duplicate_digits' requires a non-negative integer for the second argument" )

    return appendDigits( validateRealInt( n ), fmod( n, power( 10, nint( floor( k ) ) ) ), k )


#******************************************************************************
#
#  duplicateNumber
#
#******************************************************************************

@twoArgFunctionEvaluator( )
def duplicateNumber( n, k ):
    if validateRealInt( k ) == 0:
        return 0

    if k < 0:
        raise ValueError( "'duplicate_number' requires a non-negative integer for the second argument" )

    return mpmathify( getMPFIntegerAsString( n ) * int( k ) )


#******************************************************************************
#
#  reverseDigits
#
#******************************************************************************

def reverseDigits( n ):
    return mpmathify( getMPFIntegerAsString( n )[ : : -1 ] )

@oneArgFunctionEvaluator( )
def reverseDigitsOperator( n ):
    return reverseDigits( n )


#******************************************************************************
#
#  isPalindrome
#
#******************************************************************************

def isPalindrome( n ):
    result = getMPFIntegerAsString( n )

    length = len( result )

    for i in range( 0, length // 2 ):
        if result[ i ] != result[ -( i + 1 ) ]:
            return 0

    return 1

@oneArgFunctionEvaluator( )
def isPalindromeOperator( n ):
    return isPalindrome( n )


#******************************************************************************
#
#  isPandigital
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def isPandigital( n ):
    n = getMPFIntegerAsString( n )

    length = len( n )

    if length < 10:
        return 0

    digitsToCheck = string.digits

    for c in digitsToCheck:
        if c not in n:
            return 0

    return 1


#******************************************************************************
#
#  containsDigits
#
#******************************************************************************

@twoArgFunctionEvaluator( )
def containsDigits( n, k ):
    n = getMPFIntegerAsString( n )

    if isinstance( k, ( mpf, int, float ) ):
        k = getMPFIntegerAsString( k )

    for c in k:
        if c not in n:
            return 0

    return 1


#******************************************************************************
#
#  containsAnyDigits
#
#******************************************************************************

@twoArgFunctionEvaluator( )
def containsAnyDigits( n, k ):
    n = getMPFIntegerAsString( n )

    if isinstance( k, ( mpf, int, float ) ):
        k = getMPFIntegerAsString( k )

    for c in k:
        if c in n:
            return 1

    return 0


#******************************************************************************
#
#  containsOnlyDigits
#
#******************************************************************************

@twoArgFunctionEvaluator( )
def containsOnlyDigits( n, k ):
    if isinstance( k, ( mpf, int, float ) ):
        k = getMPFIntegerAsString( k )

    for c in getMPFIntegerAsString( n ):
        if c not in k:
            return 0

    return 1


#******************************************************************************
#
#  countDigits
#
#******************************************************************************

@twoArgFunctionEvaluator( )
def countDigits( n, k ):
    n = getMPFIntegerAsString( n )

    if isinstance( k, ( mpf, int, float ) ):
        k = getMPFIntegerAsString( k )

    result = 0

    for c in k:
        result += n.count( c )

    return result


#******************************************************************************
#
#  getDigitCount
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getDigitCount( n ):
    return len( getMPFIntegerAsString( n ) )


#******************************************************************************
#
#  countDifferentDigits
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def countDifferentDigits( n ):
    return len( list( set( getMPFIntegerAsString( n ) ) ) )


#******************************************************************************
#
#  getNthReversalAdditionGenerator
#
#  https://en.wikipedia.org/wiki/Lychrel_number
#
#******************************************************************************

@twoArgFunctionEvaluator( )
def getNthReversalAdditionGenerator( n, k ):
    next = int( validateRealInt( n ) )
    yield next

    for _ in arange( k ):
        next = fadd( reverseDigits( next ), next )

        yield next

        if isPalindrome( next ):
            break

def getNthReversalAddition( n, k ):
    return RPNGenerator( getNthReversalAdditionGenerator( n, k ) )


#******************************************************************************
#
#  findPalindrome
#
#******************************************************************************

@twoArgFunctionEvaluator( )
def findPalindrome( n, k ):
    next = int( validateRealInt( n ) )

    for i in range( int( validateRealInt( k ) ) + 1 ):
        next = reverseDigits( next ) + next

        if isPalindrome( next ):
            return [ i + 1, next ]

    return [ k, 0 ]


#******************************************************************************
#
#  isNarcissistic
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def isNarcissistic( n ):
    digits = getDigitList( validateRealInt( n ) )

    count = len( digits )

    sum = 0

    for digit in digits:
        sum = fadd( sum, power( digit, count ) )

        if sum > n:
            return 0

    return 1 if sum == n else 0


#******************************************************************************
#
#  isPerfectDigitalInvariant
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def isPerfectDigitalInvariant( n ):
    digits = getDigitList( validateRealInt( n ) )

    exponent = 1

    oldsum = 0

    while True:
        sum = 0

        for digit in digits:
            sum = fadd( sum, power( digit, exponent ) )

            if sum > n:
                return 0
            elif sum == n:
                return 1

        if sum == oldsum:
            return 0

        oldsum = sum

        exponent += 1


#******************************************************************************
#
#  isBaseKNarcissistic
#
#******************************************************************************

@twoArgFunctionEvaluator( )
def isBaseKNarcissistic( n, k ):
    digits = getBaseKDigits( validateRealInt( n ), k )

    count = len( digits )

    sum = 0

    for digit in digits:
        sum = fadd( sum, power( digit, count ) )

        if sum > n:
            return 0

    return 1 if sum == n else 0


#******************************************************************************
#
#  isGeneralizedDudeneyNumber
#
#  http://www.jakob.at/steffen/dudeney.html
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@cachedFunction( 'generalized_dudeney' )
def isGeneralizedDudeneyNumber( base, exponent ):
    precision = fadd( fmul( base, exponent ), 2 )

    if mp.dps < precision:
        mp.dps = precision

    n = power( validateRealInt( base ), validateRealInt( exponent ) )
    return 1 if sumDigits( n ) == base else 0


#******************************************************************************
#
#  isPerfectDigitToDigitInvariant
#
#  https://en.wikipedia.org/wiki/Perfect_digit-to-digit_invariant
#
#******************************************************************************

@twoArgFunctionEvaluator( )
def isPerfectDigitToDigitInvariant( n, k ):
    digits = getBaseKDigits( validateRealInt( n ), k )

    sum = 0

    for digit in digits:
        digit = int( digit )

        if digit != 0:
            sum = fadd( sum, power( digit, digit ) )

        if sum > n:
            return 0

    return 1 if sum == n else 0


#******************************************************************************
#
#  isSumProductNumber
#
#  https://en.wikipedia.org/wiki/Sum-product_number
#
#******************************************************************************

@twoArgFunctionEvaluator( )
def isSumProductNumber( n, k ):
    digits = getBaseKDigits( validateRealInt( n ), k )
    sum = fmul( fsum( digits ), fprod( digits ) )
    return 1 if sum == n else 0


#******************************************************************************
#
#  isHarshadNumber
#
#  https://en.wikipedia.org/wiki/Harshad_number
#
#******************************************************************************

@twoArgFunctionEvaluator( )
def isHarshadNumber( n, k ):
    digits = getBaseKDigits( validateRealInt( n ), k )
    return 1 if isDivisible( n, fsum( digits ) ) else 0


#******************************************************************************
#
#  isKaprekarNumber
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def isKaprekarNumber( n ):
    if n == 1:
        return 1
    elif n < 9:
        return 0

    a, b = splitNumberByDigits( power( n, 2 ) )

    return 1 if ( fadd( a, b ) == n ) else 0


#******************************************************************************
#
#  convertStringsToNumbers
#
#******************************************************************************

def convertStringsToNumbers( values ):
    for i in values:
        yield mpmathify( i )


#******************************************************************************
#
#  buildNumbers
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def buildNumbers( expression ):
    digitLists = parseNumbersExpression( expression )

    if len( digitLists ) == 1:
        return RPNGenerator.createGenerator( convertStringsToNumbers, digitLists )
    else:
        return RPNGenerator.createStringProduct( digitLists )


#******************************************************************************
#
#  buildLimitedDigitNumbers
#
#******************************************************************************

def buildLimitedDigitNumbers( digits, minLength, maxLength ):
    if minLength < 1:
        raise ValueError( 'minimum length must be greater than 0' )

    if  maxLength < minLength:
        raise ValueError( 'maximum length must be greater than or equal to minimum length' )

    for i in range( minLength, maxLength + 1 ):
        for item in itertools.product( digits, repeat=i ):
            number = ''.join( item )

            yield number


#******************************************************************************
#
#  parseNumbersExpression
#
#  D[D]...
#
#  One or more digit expressions where each digit expression can be:
#      - A single digit, "0" through "9"
#      - "d" - digit (equivalent to "[0-9]"
#      - "[I[I]...]" - any of the individual digit expressions
#      - "[I[I]...:m]" - permutations of any digits I, m digits long
#      - "[I[I]...:n:m]" - permutations of any digits I, from a minimum of n,
#                          up to a maximum of m digits long
#
#  A digit expression (I) can be a single digit or a range.  A single digit is
#  anything from "0" through "9".  A digit range looks like: "a-b" where a
#  is a digit that is smaller than b.
#
#******************************************************************************

def parseNumbersExpression( arg ):
    if not isinstance( arg, str ):
        arg = str( validateRealInt( arg ) )

    result = [ ]

    digitRange = [ '0', '1', '2', '3', '4', '5', '6', '7', '8', '9' ]
    oddRange = [ '1', '3', '5', '7', '9' ]
    evenRange = [ '0', '2', '4', '6', '8' ]

    defaultState = 0
    startDigitState = 1
    digitState = 2
    startRangeState = 3
    numberLengthRange1 = 4
    numberLengthRange2 = 5

    state = defaultState

    currentGroup = set( )

    lengthRangeLow = ''
    lengthRangeHigh = ''

    for char in arg:
        #print( 'state', state, 'ch', ch, 'result', result, 'currentGroup', currentGroup )
        if state == defaultState:
            if char == 'd':
                result.append( digitRange )
            elif char == 'e':
                result.append( evenRange )
            elif char == 'o':
                result.append( oddRange )
            elif char == '[':
                state = digitState
                currentGroup = set( )
            elif '0' <= char <= '9':
                result.append( [ char ] )
            else:
                raise ValueError( 'unexpected character \'{}\''.format( char ) )
        elif state == startDigitState:
            if '0' <= char <= '9':
                currentGroup.add( char )
                state = digitState
                lastDigit = char
            elif char == ']':
                result.append( sorted( list( currentGroup ) ) )
                state = defaultState
            else:
                raise ValueError( 'unexpected character \'{}\''.format( char ) )
        elif state == digitState:
            if '0' <= char <= '9':
                currentGroup.add( char )
                lastDigit = char
            elif char == 'd':
                for digit in digitRange:
                    currentGroup.add( digit )

                lastDigit = ' '
            elif char == 'e':
                for digit in evenRange:
                    currentGroup.add( digit )

                lastDigit = ' '
            elif char == 'o':
                for digit in oddRange:
                    currentGroup.add( digit )

                lastDigit = ' '
            elif char == '-':
                state = startRangeState
            elif char == ':':
                state = numberLengthRange1
            elif char == ']':
                result.append( sorted( list( currentGroup ) ) )
                state = defaultState
        elif state == startRangeState:
            if lastDigit == ' ':
                raise ValueError( 'invalid digit range' )

            if '0' <= char <= '9':
                if char <= lastDigit:
                    raise ValueError( 'invalid digit range' )

                for i in range( int( lastDigit ), int( char ) + 1 ):
                    currentGroup.add( str( i ) )

                state = digitState
        elif state == numberLengthRange1:
            if '0' <= char <= '9':
                lengthRangeHigh += char
            elif char == ':':
                lengthRangeLow = lengthRangeHigh
                lengthRangeHigh = ''
                state = numberLengthRange2
            elif char == ']':
                result.append( buildLimitedDigitNumbers( sorted( list( currentGroup ) ),
                                                         int( lengthRangeHigh ), int( lengthRangeHigh ) ) )
                lengthRangeLow = ''
                lengthRangeHigh = ''
                state = defaultState
            else:
                raise ValueError( 'unexpected character \'{}\''.format( char ) )
        elif state == numberLengthRange2:
            if '0' <= char <= '9':
                lengthRangeHigh += char
            elif char == ']':
                result.append( buildLimitedDigitNumbers( sorted( list( currentGroup ) ),
                                                         int( lengthRangeLow ), int( lengthRangeHigh ) ) )
                lengthRangeLow = ''
                lengthRangeHigh = ''
                state = defaultState
            else:
                raise ValueError( 'unexpected character \'{}\''.format( char ) )

    return result


#******************************************************************************
#
#  hasUniqueDigits
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def hasUniqueDigits( n ):
    digits = getDigitList( validateRealInt( n ) )

    existing = set( )

    for digit in digits:
        if digit in existing:
            return 0
        else:
            existing.add( digit )

    return 1


#******************************************************************************
#
#  hasDigits
#
#******************************************************************************

#@twoArgFunctionEvaluator( )
#def hasDigits( value, digits ):
#    pass


#******************************************************************************
#
#  isKMorphic
#
#  This code won't work correctly for integral powers of 10, but they can
#  never be morphic anyway, except for 1, which I handle specially.
#
#******************************************************************************

def isKMorphic( n, k ):
    '''
    Returns true if n to the k power ends with n.
    '''
    if n == 1:
        return 1

    modulo = power( 10, ceil( log10( n ) ) )
    powmod = getPowMod( n, validateRealInt( k ), modulo )

    return 1 if ( n == powmod ) else 0

@twoArgFunctionEvaluator( )
def isKMorphicOperator( n, k ):
    return isKMorphic( n, k )

@oneArgFunctionEvaluator( )
def isAutomorphic( n ):
    return isKMorphic( n, 2 )

@oneArgFunctionEvaluator( )
def isTrimorphic( n ):
    return isKMorphic( n, 3 )


#******************************************************************************
#
#  getLeftTruncations
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getLeftTruncations( n ):
    if n < 0:
        raise ValueError( '\'get_left_truncations\' requires a positive argument' )

    n = getMPFIntegerAsString( n )

    for i, _ in enumerate( n ):
        yield mpmathify( n[ i : ] )

@oneArgFunctionEvaluator( )
def getLeftTruncationsGenerator( n ):
    return RPNGenerator.createGenerator( getLeftTruncations, n )


#******************************************************************************
#
#  getRightTruncations
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getRightTruncations( n ):
    if n < 0:
        raise ValueError( '\'get_right_truncations\' requires a positive argument' )

    result = getMPFIntegerAsString( n )

    for i in range( len( result ), 0, -1 ):
        yield mpmathify( result[ 0 : i ] )

@oneArgFunctionEvaluator( )
def getRightTruncationsGenerator( n ):
    return RPNGenerator.createGenerator( getRightTruncations, n )


#******************************************************************************
#
#  getMultiplicativePersistence
#
#******************************************************************************

def getMultiplicativePersistence( n, exponent = 1, dropZeroes = False, persistence = 0 ):
    if exponent == 1 and n < 10:
        return persistence

    if n <= 1:
        return persistence
    else:
        return getMultiplicativePersistence( multiplyDigitList( n, exponent, dropZeroes ),
                                             exponent, dropZeroes, persistence + 1 )

@oneArgFunctionEvaluator( )
def getPersistence( n ):
    return getMultiplicativePersistence( n )

@twoArgFunctionEvaluator( )
def getKPersistence( n, k ):
    return getMultiplicativePersistence( n, k )

@oneArgFunctionEvaluator( )
def getErdosPersistence( n ):
    return getMultiplicativePersistence( n, 1, True )



#******************************************************************************
#
#  showMultiplicativePersistence
#
#******************************************************************************

def showMultiplicativePersistence( n, exponent = 1, dropZeroes = False ):
    yield n

    if exponent == 1:
        while n >= 10:
            n = multiplyDigitList( n, exponent, dropZeroes )
            yield n
    else:
        while n > 1:
            n = multiplyDigitList( n, exponent, dropZeroes )
            yield n

@oneArgFunctionEvaluator( )
def showPersistence( n ):
    return RPNGenerator.createGenerator( showMultiplicativePersistence, [ n ] )

@twoArgFunctionEvaluator( )
def showKPersistence( n, k ):
    return RPNGenerator.createGenerator( showMultiplicativePersistence, [ n, k ] )


#******************************************************************************
#
#  showErdosPersistenceGenerator
#
#******************************************************************************

def showErdosPersistenceGenerator( n ):
    yield n

    while n >= 10:
        n = multiplyDigitList( n, 1, True )
        yield n

@oneArgFunctionEvaluator( )
def showErdosPersistence( n ):
    return RPNGenerator.createGenerator( showErdosPersistenceGenerator, [ n ] )

@oneArgFunctionEvaluator( )
def permuteDigits( n ):
    return RPNGenerator.createPermutations( getMPFIntegerAsString( n ) )


#******************************************************************************
#
#  isIncreasing
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def isIncreasing( n ):
    n = getMPFIntegerAsString( n )

    for i in range( 1, len( n ) ):
        if n[ i ] < n[ i - 1 ]:
            return 0

    return 1


#******************************************************************************
#
#  isDecreasing
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def isDecreasing( n ):
    n = getMPFIntegerAsString( n )

    for i in range( 1, len( n ) ):
        if n[ i ] > n[ i - 1 ]:
            return 0

    return 1


#******************************************************************************
#
#  isBouncy
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def isBouncy( n ):
    if isIncreasing( n ) == 0 and isDecreasing( n ) == 0:
        return 1
    else:
        return 0


#******************************************************************************
#
#  rotateDigitsLeft
#
#******************************************************************************

@twoArgFunctionEvaluator( )
def rotateDigitsLeft( n, k ):
    if k < 0:
        return rotateDigitsRight( n, fneg( k ) )

    n = getMPFIntegerAsString( n )

    if k > len( n ):
        raise ValueError( 'cannot rotate more digits than the number has' )

    rotate = int( k )

    n = n[ rotate : ] + n[ : rotate ]
    return mpmathify( n )


#******************************************************************************
#
#  rotateDigitsRight
#
#******************************************************************************

@twoArgFunctionEvaluator( )
def rotateDigitsRight( n, k ):
    if k < 0:
        return rotateDigitsLeft( n, fneg( k ) )

    n = getMPFIntegerAsString( n )

    if k > len( n ):
        raise ValueError( 'cannot rotate more digits than the number has' )

    rotate = int( k )

    n = n[ -rotate : ] + n[ : -rotate ]
    return mpmathify( n )


#******************************************************************************
#
#  getCyclicPermutations
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getCyclicPermutations( n ):
    result = [ n ]

    n = getMPFIntegerAsString( n )

    for _ in range( len( n ) - 1 ):
        n = n[ 1 : ] + n[ : 1 ]
        result.append( mpmathify( n ) )

    return result


#******************************************************************************
#
#  isDigitalPermutation
#
#******************************************************************************

@twoArgFunctionEvaluator( )
def isDigitalPermutation( n, k ):
    str1 = getMPFIntegerAsString( n )
    str2 = getMPFIntegerAsString( k )

    if len( str1 ) != len( str2 ):
        return 0

    if sorted( str1 ) != sorted( str2 ):
        return 0

    return 1


#******************************************************************************
#
#  getSquareDigitChainGenerator
#
#******************************************************************************

def generateSquareDigitChainGenerator( n ):
    n = validateRealInt( floor( n ) )

    chain = [ ]

    if n == 0:
        yield 0
        return

    if n == 1:
        yield 1
        return

    done = False

    while not done:
        digits = getDigits( n )

        n = 0

        for i in digits:
            n = fadd( n, pow( i, 2 ) )

        if n in chain:
            done = True
        else:
            chain.append( n )

            yield n

@oneArgFunctionEvaluator( )
def generateSquareDigitChain( n ):
    return RPNGenerator.createGenerator( generateSquareDigitChainGenerator, [ n ] )


#******************************************************************************
#
#  isStepNumber
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def isStepNumber( n ):
    if n < 10:
        return 0

    n = getMPFIntegerAsString( n )

    for i in range( 1, len( n ) ):
        if abs( int( n[ i ] ) - int( n[ i - 1 ] ) ) != 1:
            return 0

    return 1


#******************************************************************************
#
#  buildNextStepNumbers
#
#  This method expects a list of digits.
#
#******************************************************************************

def buildNextStepNumbers( n ):
    result = [ ]

    if n[ -1 ] == 0:
        result.append( n + [ 1 ] )
    elif n[ -1 ] == 9:
        result.append( n + [ 8 ] )
    else:
        result.append( n + [ n[ -1 ] - 1 ] )
        result.append( n + [ n[ -1 ] + 1 ] )

    return result


#******************************************************************************
#
#  buildStepNumbers
#
#******************************************************************************

def buildStepNumbersGenerator( maxLength ):
    if maxLength < 2:
        raise ValueError( '\'build_step_numbers\' requires an argument of 2 or greater' )

    stepNumbers = [ ]

    for i in range( 1, 10 ):
        stepNumbers.extend( buildNextStepNumbers( [ i ] ) )

    for i in stepNumbers:
        yield combineDigits( i )

    for i in range( 0, int( maxLength ) - 2 ):
        newStepNumbers = [ ]

        for j in stepNumbers:
            newStepNumbers.extend( buildNextStepNumbers( j ) )

        stepNumbers = newStepNumbers

        for j in stepNumbers:
            yield combineDigits( j )

@oneArgFunctionEvaluator( )
def buildStepNumbers( n ):
    return RPNGenerator.createGenerator( buildStepNumbersGenerator, [ n ] )


#******************************************************************************
#
#  isSmithNumber
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@cachedFunction( 'smith' )
def isSmithNumber( n ):
    if isPrime( validateRealInt( n ) ) or n < 2:
        return 0

    sum1 = sumDigits( n )
    sum2 = fsum( [ sumDigits( i ) for i in getFactors( n ) ] )

    return 1 if sum1 == sum2 else 0


#******************************************************************************
#
#  isBaseKSmithNumber
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@cachedFunction( 'base_k_smith' )
def isBaseKSmithNumber( n, k ):
    if validateRealInt( k ) < 2:
        raise ValueError( "'is_base_k_smith_number' base argument must be 2 or greater" )

    if k > 4 and n == 4:
        return 1

    if k > 4 and n < k:
        return 0

    if isPrime( validateRealInt( n ) ) or n < 2:
        return 0

    sum1 = fsum( getBaseKDigits( n, k ) )
    sum2 = fsum( [ fsum( getBaseKDigits( i, k ) ) for i in getFactors( n ) ] )

    return 1 if sum1 == sum2 else 0


#******************************************************************************
#
#  isOrderKSmithNumber
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@cachedFunction( 'order_k_smith' )
def isOrderKSmithNumber( n, k ):
    if isPrime( validateRealInt( n ) ) or n < 2:
        return 0

    digitList1 = getNonzeroDigits( n )
    digitList2 = [ item for sublist in [ getNonzeroDigits( m ) for m in getFactors( n ) ] for item in sublist ]

    if sorted( digitList1 ) == sorted( digitList2 ):
        return 0

    sum1 = fsum( [ power( i, k ) for i in getDigits( n ) ] )
    sum2 = fsum( [ power( j, k ) for j in [ item for sublist in \
                        [ getDigits( m ) for m in getFactors( n ) ] for item in sublist ] ] )

    return 1 if sum1 == sum2 else 0


#  Look and Say from:
#  https://www.rosettacode.org/wiki/Look-and-say_sequence#Python

#>>> from itertools import groupby
#>>> def lookandsay(number):
#    return ''.join( str(len(list(g))) + k
#                for k,g in groupby(number) )
#
#>>> numberstring='1'
#>>> for i in range(10):
#    print numberstring
#    numberstring = lookandsay(numberstring)
#



#!/usr/bin/env python

#******************************************************************************
#
#  rpnInput.py
#
#  rpnChilada input functions
#  copyright (c) 2020, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

import math

import arrow

from mpmath import fdiv, fneg, mp, mpc, mpf, mpmathify

from rpn.rpnDateTime import RPNDateTime
from rpn.rpnGenerator import RPNGenerator
from rpn.rpnSettings import setAccuracy
from rpn.rpnUtils import oneArgFunctionEvaluator

import rpn.rpnGlobals as g


#******************************************************************************
#
#  convertToBase10
#
#******************************************************************************

def convertToBase10( integer, mantissa, inputRadix ):
    result = mpmathify( 0 )
    base = mpmathify( 1 )

    validNumerals = g.numerals[ : inputRadix ]

    for i in range( len( integer ) - 1, -1, -1 ):
        digit = validNumerals.find( integer[ i ] )

        if digit == -1:
            raise ValueError( 'invalid numeral \'%c\' for base %d' % ( integer[ i ], inputRadix ) )

        result += digit * base
        base *= inputRadix

    base = fdiv( 1, inputRadix )

    for i, numeral in enumerate( mantissa ):
        digit = validNumerals.find( numeral )

        if digit == -1:
            raise ValueError( 'invalid numeral \'%c\' for base %d' % ( mantissa[ i ], inputRadix ) )

        result += digit * base
        base /= inputRadix

    return result


#******************************************************************************
#
#  parseInputValue
#
#  Parse out a datetime value or a numerical expression and attempt to set
#  the precision to an appropriate value based on the expression.
#
#******************************************************************************

def parseInputValue( term, inputRadix = 10 ):
    if isinstance( term, mpf ):
        return term

    possibleDate = True

    tCount = 0

    # one 'T' is allowed in a date, but no other letters
    for c in term:
        if c.isalpha( ):
            if c == 'T':
                if tCount > 0:
                    possibleDate = False
                    break

                tCount += 1
            else:
                possibleDate = False
                break

    if not g.interactive:
        if term[ 0 ] == '$' and term[ 1 : ] in g.userVariables:
            term = g.userVariables[ term[ 1 : ] ]
            return term

        if term[ 0 ] == '@' and term[ 1 : ] in g.userFunctions:
            term = g.userFunctions[ term[ 1 : ] ]
            return term

        if term[ 0 ] == "'":
            return term[ 1 : ]

    innerChars = term[ 1 : -1 ]

    # this helps us parse dates
    if '/' in innerChars and possibleDate:
        term = term.replace( '/', '-' )
        innerChars = term[ 1 : -1 ]

    # 'e' implies scientific notation, which isn't a date regardless
    if possibleDate:
        # 'd' means a dice expression, '[' means a build_number expression, so don't treat it as a date
        if ( ( '-' in innerChars ) or ( ':' in innerChars ) ) and ( '[' not in term ):
            # try:
            datetime = arrow.get( term )
            datetime = RPNDateTime( datetime.year, datetime.month, datetime.day,
                                    datetime.hour, datetime.minute, datetime.second,
                                    datetime.microsecond )
            # except:
            #     raise ValueError( 'error parsing datetime' )

            return datetime

    if term == '0':
        return mpmathify( 0 )

    # ignore commas
    term = ''.join( [ i for i in term if i not in ',' ] )

    if term[ 0 ] == '\\':
        term = term[ 1 : ]
        ignoreSpecial = True
    else:
        ignoreSpecial = False

    if inputRadix == 10:
        imaginary = term[ -1 ] in ( 'i', 'j' )
    else:
        imaginary = False

    if imaginary:
        term = term[ : -1 ]

    if '.' in term:
        if inputRadix == 10:
            newPrecision = len( term ) + 1

            if mp.dps < newPrecision:
                setAccuracy( newPrecision )

            if imaginary:
                return mpc( imag = term )
            else:
                return mpmathify( term )

        decimal = term.find( '.' )
    else:
        decimal = len( term )

    negative = term[ 0 ] == '-'

    if negative:
        start = 1
    else:
        if term[ 0 ] == '+':
            start = 1
        else:
            start = 0

    integer = term[ start : decimal ]
    mantissa = term[ decimal + 1 : ]

    # check for hex, then binary, then octal, otherwise a plain old decimal integer
    if not ignoreSpecial and mantissa == '':
        if integer[ 0 ] == '0':
            if len( integer ) == 1:
                return mpmathify( 0 )

            if integer[ 1 ] in 'Xx':
                # set the precision big enough to handle this value
                newPrecision = math.ceil( ( math.log10( 16 ) * ( len( integer ) - 2 ) ) ) + 1

                if mp.dps < newPrecision:
                    setAccuracy( newPrecision )

                return mpmathify( int( integer[ 2 : ], 16 ) )
            elif integer[ -1 ] in 'bB':
                # set the precision big enough to handle this value
                newPrecision = math.ceil( math.log10( 2 ) * ( len( integer ) - 1 ) ) + 1

                if mp.dps < newPrecision:
                    setAccuracy( newPrecision )

                integer = integer[ : -1 ]
                return mpmathify( int( integer, 2 ) * ( -1 if negative else 1 ) )
            else:
                integer = integer[ 1 : ]

                return mpmathify( int( integer, 8 ) )
        if integer[ 0 ] == '1' and integer[ -1 ] in 'bB':
            # set the precision big enough to handle this value
            newPrecision = math.ceil( math.log10( 2 ) * ( len( integer ) - 1 ) ) + 1

            if mp.dps < newPrecision:
                setAccuracy( newPrecision )

            integer = integer[ : -1 ]
            return mpmathify( int( integer, 2 ) * ( -1 if negative else 1 ) )
        elif inputRadix == 10:
            newPrecision = len( integer ) + 1

            if mp.dps < newPrecision:
                setAccuracy( newPrecision )

            result = fneg( integer ) if negative else mpmathify( integer )

            if imaginary:
                return mpc( real = '0.0', imag = result )
            else:
                return result

    # finally, we have a non-radix 10 number with a mantissa to parse
    result = convertToBase10( integer, mantissa, inputRadix )
    result = fneg( result ) if negative else mpmathify( result )

    if imaginary:
        return mpc( real = '0.0', imag = result )
    else:
        return result


#******************************************************************************
#
#  readListFromFile
#
#******************************************************************************

def readListFromFileGenerator( filename ):

    with open( filename ) as file:
        for i in file:
            if i == '\n':
                continue

            try:
                yield parseInputValue( i )
            except:
                pass

@oneArgFunctionEvaluator( )
def readListFromFile( filename ):
    return RPNGenerator( readListFromFileGenerator( filename ) )




#!/usr/bin/env python

#******************************************************************************
#
#  rpnConstantOperators.py
#
#  rpnChilada constant operator declarations
#  copyright (c) 2020, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

from mpmath import fdiv, fmul, mpmathify, pi

from rpn.rpnConstantUtils import getMaxDouble, getMaxFloat, getMinDouble, getMinFloat
from rpn.rpnUnitClasses import RPNConstantInfo


#******************************************************************************
#
#  constantOperators
#
#  unit name : value, unit, aliases, multipliable, description
#
#  When unit types are multiplied in compound units, they need to be
#  specified in alphabetical order in the name, but not the representations.
#
#******************************************************************************

constantOperators = {
    # physical constants
    'avogadro_number' :
        RPNConstantInfo( '6.022140756e23', '',
                         [ 'avogadro', 'avogadros_number', 'N_sub_A' ], False,
                         'Avogadro\'s number, the number of atoms in a mole',
                         '''
Ref:  https://www.bipm.org/utils/en/pdf/si-revised-brochure/Draft-SI-Brochure-2018.pdf
''' ),

    'boltzmann_constant' :
        RPNConstantInfo( '1.380649e-23', 'kilogram*meter^2/second^2*kelvin',
                         [ 'boltzmann', 'boltzmanns_const', 'k_sub_b', 'k_b' ], False,
                         'The Boltzmann constant relates the average kinetic energy of particles '
                         'in a gas to the temperature of the gas',
                         '''
Ref:  https://en.wikipedia.org/wiki/Boltzmann_constant
      https://www.bipm.org/utils/en/pdf/si-revised-brochure/Draft-SI-Brochure-2018.pdf

This value is now exact, by definition.
''' ),

    'bohr_radius' :
        RPNConstantInfo( '5.29177210903e-11', 'meter', [ ], False,
                         'the Bohr radius, an idealized measure of the size of a hydrogen atom',
                         '''
From https://en.wikipedia.org/wiki/Bohr_radius:

The Bohr radius (a0) is a physical constant, equal to the most probable
distance between the nucleus and the electron in a hydrogen atom in its ground
state (non-relativistic and with an infinitely heavy proton).  It is named
after Niels Bohr, due to its role in the Bohr model of an atom.

Ref:  CODATA 2014 value - https://physics.nist.gov/cgi-bin/cuu/Value?bohrrada0
''' ),

    'classical_electron_radius' :
        RPNConstantInfo( '2.8179403227e-15', 'meter', [ 'electron_radius' ], False,
                         'the classical size of an electron',
                         '''
Ref:  https://en.wikipedia.org/wiki/Classical_electron_radius
''' ),

    'coulomb_constant' :
        RPNConstantInfo( '8.9875517873681764e9', 'kilogram*meter^3/ampere^2*second^4', [ ], False,
                         'Coulomb\'s constant',
                         '''
Ref:  https://en.wikipedia.org/wiki/Coulomb%27s_constant
''' ),

    'electric_constant' :
        RPNConstantInfo( '8.8541878128e-12', 'ampere^2*second^4/kilogram*meter^3',
                         [ 'e0', 'e_0', 'e_nought', 'e_sub_0', 'free_space_permittivity',
                           'vacuum_permittivity' ], False,
                         'the permittivity of free space',
                         '''
The physical constant e_0 (pronounced as "epsilon nought" or "epsilon zero"),
commonly called the vacuum permittivity, permittivity of free space or
dielectric constant or the distributed capacitance of the vacuum, is an ideal,
(baseline) physical constant, which is the value of the absolute dielectric
permittivity of classical vacuum.

It is the capability of an electric field to permeate a vacuum.  This constant
relates the units for electric charge to mechanical quantities such as length
and force.

Ref:  https://en.wikipedia.org/wiki/Vacuum_permittivity
''' ),

    'electron_charge' :
        RPNConstantInfo( '1.602176634e-19', 'coulomb', [ 'elementary_charge', 'electron' ], False,
                         'the measured charge of an electron',
                         '''
Ref:  https://en.wikipedia.org/wiki/2019_redefinition_of_SI_base_units#Ampere

This new value is exact, by definition.
''' ),

    'hyperfine_transition_frequency_of_cesium' :
        RPNConstantInfo( '9192631770', 'Hz', [ 'delta_nu_sub_cs' ], False,
                         'the frequency of the ground state transitions of the cesium atom',
                         '''
Since 1967, the second has been defined as exactly "the duration of
9,192,631,770 periods of the radiation corresponding to the transition between
the two hyperfine levels of the ground state of the caesium-133 atom" (at a
temperature of 0 K).  This length of a second was selected to correspond
exactly to the length of the ephemeris second previously defined.  Atomic
clocks use such a frequency to measure seconds by counting cycles per second
at that frequency.  Radiation of this kind is one of the most stable and
reproducible phenomena of nature. The current generation of atomic clocks are
accurate to within one second in a few hundred million years.

Ref:  https://en.wikipedia.org/wiki/Second#%22Atomic%22_second
Ref:  https://www.bipm.org/utils/en/pdf/si-revised-brochure/Draft-SI-Brochure-2018.pdf
''' ),

    'magnetic_constant' :
        RPNConstantInfo( '1.25663706212e-6', 'newton/ampere^2',
                         [ 'free_space_permeability', 'mu0', 'mu_0', 'mu_sub_0', 'mu_nought' ], False,
                         'the permeability of free space',
                         '''
The physical constant, ('mu_0'), commonly called the vacuum permeability,
permeability of free space, permeability of vacuum, or magnetic constant, is
the magnetic permeability in a classical vacuum.  Vacuum permeability is
derived from production of a magnetic field by an electric current or by a
moving electric charge and in all other formulas for magnetic-field production
in a vacuum.

Ref:  https://en.wikipedia.org/wiki/Vacuum_permeability
''' ),

    'magnetic_flux_quantum' :
        RPNConstantInfo( '2.067833831e-15', 'weber',
                         [ 'magnetic_flux_quanta', 'josephson_constant', 'K_sub_j' ], False,
                         'the magnetic flux quantum',
                         '''
From https://en.wikipedia.org/wiki/Magnetic_flux_quantum:

The (superconducting) magnetic flux quantum F0 = h/2e =~
2.067833831(13) x 10e-15 Wb is a combination of fundamental physical constants:
the Planck constant h and the electron charge e.  Its value is, therefore, the
same for any superconductor.

Ref:  CODATA 2014 value - https://physics.nist.gov/cgi-bin/cuu/Value?flxquhs2e
''' ),

    'molar_gas_constant' :
        RPNConstantInfo( '8.3144598', 'joule/mole*kelvin',
                         [ 'gas_constant', 'ideal_gas_constant', 'universal_gas_constant' ], False,
                         'molar gas constant',
                         '''
From https://en.wikipedia.org/wiki/Gas_constant:

The gas constant (also known as the molar, universal, or ideal gas constant)
is denoted by the symbol R.  It is equivalent to the Boltzmann constant, but
expressed in units of energy per temperature increment per mole, i.e. the
pressure-volume product, rather than energy per temperature increment per
particle.  The constant is also a combination of the constants from Boyle's
law, Charles's law, Avogadro's law, and Gay-Lussac's law.  It is a physical
constant that is featured in many fundamental equations in the physical
sciences, such as the ideal gas law, the Arrhenius equation, and the Nernst
equation.

Ref:  CODATA 2014 value - https://physics.nist.gov/cgi-bin/cuu/Value?r
''' ),

    'newton_constant' :
        RPNConstantInfo( '6.67408e-11', 'meter^3/kilogram*second^2', [ 'G', 'newtons_constant' ], False,
                         'Newton\'s constant of gravitation',
                         '''
Ref:  CODATA 2014 value - https://physics.nist.gov/cgi-bin/cuu/Value?bg
''' ),

    'nuclear_magneton' :
        RPNConstantInfo( '5.050783699e-27', 'joule/tesla', [ ], False,
                         'the nuclear magneton, a physical constant of magnetic moment',
                         '''
From https://en.wikipedia.org/wiki/Nuclear_magneton:

Ref:  CODATA 2014 value - https://physics.nist.gov/cgi-bin/cuu/Value?mun

''' ),

    'planck_constant' :
        RPNConstantInfo( '6.62607015e-34', 'kilogram*meter^2/second', [ 'h', 'planck' ], False,
                         'Planck\'s constant, the quantum of electromagnetic action that relates a photon\'s energy to its frequency',
                         '''
From:  https://en.wikipedia.org/wiki/Planck_constant

The Planck constant, or Planck's constant, is the quantum of electromagnetic
action that relates a photon's energy to its frequency.  The Planck constant
multiplied by a photon's frequency is equal to a photon's energy.  The Planck
constant is a fundamental physical constant denoted as h, and of fundamental
importance in quantum mechanics.  In metrology it is used to define the
kilogram in SI units.

The Planck constant is defined to have the exact value h=6.62607015 x 10e-34
joule-second in SI units.

This is the exact value, set on 20 Nov 2018 by the General Conference on
Weights and Measures.
''' ),

    'reduced_planck_constant' :
        RPNConstantInfo( fdiv( mpmathify( '6.62607015e-34' ), fmul( 2, pi ) ), 'kilogram*meter^2/second',
                         [ 'h_bar', 'reduced_planck', 'dirac', 'dirac_constant' ], False,
                         'Planck\'s constant divided by pi',
                         '''
From:  https://en.wikipedia.org/wiki/Planck_constant

The Planck constant has dimensions of physical action; i.e., energy multiplied
by time, or momentum multiplied by distance, or angular momentum.  In SI units,
the Planck constant is expressed in joule-seconds.  Implicit in the dimensions
of the Planck constant is the fact that the SI unit of frequency, the Hertz,
represents one complete cycle, 360 degrees or 2 pi radians, per second.  An
angular frequency in radians per second is often more natural in mathematics
and physics and many formulas use a reduced Planck constant (pronounced h-bar).
''' ),

    'rydberg_constant' :
        RPNConstantInfo( '10973731.568508', 'meter^-1', [ ], False,
                         'the Rydberg constant, a physical constant relating to the electromagnetic spectra of an atom',
                         '''
From https://en.wikipedia.org/wiki/Rydberg_constant:


Ref:  CODATA 2014 value - https://physics.nist.gov/cgi-bin/cuu/Value?ryd
''' ),

    'speed_of_light' :
        RPNConstantInfo( '299792458', 'meter/second',
                         [ 'c', 'lightspeed', 'light_speed', 'planck_speed', 'planck_velocity', 'light' ], False,
                         'the speed of light in a vacuum',
                         '''
Ref:  CODATA 2014 value - https://physics.nist.gov/cgi-bin/cuu/Value?c
''' ),

    # physical quantities
    'aa_battery' :
        RPNConstantInfo( '15400', 'joule', [ ], True,
                         'Typical energy charge of a fully-charged alkaline AA battery',
                         '''
The is the typical energy charge of a fully-charged alkaline AA battery.
''' ),

    'gallon_of_ethanol' :
        RPNConstantInfo( '8.087e7', 'joule', [ ], True,
                         'Energy content of a U.S. gallon of ethanol',
                         '''
This unit is calculated from the relative energy content of ethanol compared to
gasoline.

https://en.wikipedia.org/wiki/Gasoline_gallon_equivalent#Gasoline_gallon_equivalent_tables
''' ),

    'gallon_of_gasoline' :
        RPNConstantInfo( '1.213e8', 'joule', [ ], True,
                         'Typical energy content of a U.S. gallon of gasoline',
                         '''
In 2011 the United States Environmental Protection Agency introduced the gallon
gasoline equivalent as a unit of energy because their research showed most
U.S. citizens do not understand the standard units.  The gallon gasoline
equivalent is defined as 33.7 kWh, or about 1.213 x 10^8 joules.

https://en.wikipedia.org/wiki/List_of_unusual_units_of_measurement#Gasoline_gallon_equivalent
https://en.wikipedia.org/wiki/Gasoline_gallon_equivalent
''' ),

    'density_of_water' :
        RPNConstantInfo( '997.0474', 'kilogram/meter^3', [ 'water' ], True,
                         'the density of pure water at 25 degress Celsius',
                         '''
The is the density of pure water at 25 degrees Celsius.
''' ),

    'solar_constant' :
        RPNConstantInfo( '1360.8', 'watt/meter^2', [ ], True,
                         'the average maximum energy received from the Sun on the Earth\'s surface',
                         '''
This constant represents the average maximum amount of luminous energy the
Earth receives from the Sun.  The solar constant does vary slightly over
time due to fluctuations in solar energy output.
''' ),

    # subatomic particle constants
    'alpha_particle_mass' :
        RPNConstantInfo( '6.644657230e-27', 'kilogram', [ ], True,
                         'the mass of an alpha particle (helium nucleus)',
                         '''
The alpha particle is the equivalent of a helium nucleus, and consists of two
protons and two neutrons.
''' ),

    'deuteron_mass' :
        RPNConstantInfo( '3.34358550157e-27', 'kilogram', [ ], True,
                         'the mass of a deuterium nucleus',
                         '''
The nucleus of deuterium is called a deuteron.  It has a mass of
2.013553212745(40) u (equal to 1875.612928(12) MeV).

The charge radius of the deuteron is 2.1413(25) fm.[21] Like the proton radius,
measurements using muonic deuterium produce a significantly smaller result:
2.12562(78) fm.[22] This is 6s less than the accepted CODATA 2014 value,
measured using electrons, and confirms the unresolved proton charge radius anomaly.

Ref:  https://en.wikipedia.org/wiki/Deuterium#Deuteron_mass_and_radius
''' ),

    'electron_mass' :
        RPNConstantInfo( '9.10938356e-31', 'kilogram', [ 'electron_rest_mass' ], True,
                        'the mass of the electron',
                         '''
The electron rest mass is the mass of a stationary electron, also known as the
invariant mass of the electron.  It is one of the fundamental constants of
physics.  It has a value of about 9.109 x 10^-31 kilograms or about 5.486 x 10^-4
atomic mass units, equivalent to an energy of about 8.187 x 10^-14 joules or about
0.5110 MeV.

https://en.wikipedia.org/wiki/Electron_rest_mass
''' ),

    'helion_mass' :
        RPNConstantInfo( '5.006412700e-27', 'kilogram', [ ], True,
                         'the mass of a Helium-3 nucleus',
                         '''
A helion is a short name for the naked nucleus of helium, a doubly positively
charged helium ion. In practice, helion refers specifically to the nucleus of
the helium-3 isotope, consisting of two protons and one neutron.  The nucleus
of the other stable isotope of helium, helium-4 isotope, which consists of two
protons and two neutrons, is specifically called an alpha particle.

This particle is emitted in the decay of tritium, an isotope of hydrogen.

https://en.wikipedia.org/wiki/Helion_(chemistry)
https://physics.nist.gov/cgi-bin/cuu/Value?mh
''' ),

    'muon_mass' :
        RPNConstantInfo( '1.883531594e-28', 'kilogram', [ ], True,
                         'the mass of a muon',
                         '''
The muon, (from the Greek letter mu, used to represent it) is an elementary
particle similar to the electron, with an electric charge of -1 e and a spin of
1/2, but with a much greater mass.  It is classified as a lepton.  As is the
case with other leptons, the muon is not believed to have any sub-structure -
that is, it is not thought to be composed of any simpler particles.

Ref:  https://en.wikipedia.org/wiki/Muon
''' ),

    'neutron_mass' :
        RPNConstantInfo( '1.674927471e-27', 'kilogram', [ ], True,
                         'the mass of a neutron',
                         '''
The neutron is a subatomic particle, with no net electric charge and a mass
slightly larger than that of a proton.  Protons and neutrons constitute the
nuclei of atoms.  Since protons and neutrons behave similarly within the
nucleus, and each has a mass of approximately one atomic mass unit, they are
both referred to as nucleons.  Their properties and interactions are described
by nuclear physics.

Ref:  https://en.wikipedia.org/wiki/Neutron
''' ),

    'proton_mass' :
        RPNConstantInfo( '1.672621898e-27', 'kilogram', [ ], True,
                         'the mass of a proton',
                         '''
A proton is a subatomic particle with a positive electric charge of +1
elementary charge and a mass slightly less than that of a neutron.  Protons
and neutrons, each with masses of approximately one atomic mass unit, are
collectively referred to as "nucleons".

One or more protons are present in the nucleus of every atom; they are a
necessary part of the nucleus.  The number of protons in the nucleus is the
defining property of an element, and is referred to as the atomic number
(represented by the symbol Z). Since each element has a unique number of
protons, each element has its own unique atomic number.

https://en.wikipedia.org/wiki/Proton
''' ),

    'tau_mass' :
        RPNConstantInfo( '3.16747e-27', 'kilogram', [ ], True,
                         'the mass of a tau particle',
                         '''
The tau, also called the tau lepton, tau particle, or tauon, is an elementary
particle similar to the electron, with negative electric charge and a spin of
1/2. Together with the electron, the muon, and the three neutrinos, it is a
lepton.

https://en.wikipedia.org/wiki/Tau_(particle)
''' ),

    'triton_mass' :
        RPNConstantInfo( '5.007356665e-27', 'kilogram', [ ], True,
                         'the mass of a tritium nucleus',
                         '''
Tritium, also known as hydrogen-3, is a radioactive isotope of hydrogen.  The
nucleus of tritium (sometimes called a triton) contains one proton and two
neutrons, whereas the nucleus of protium (by far the most abundant hydrogen
isotope) contains one proton and no neutrons.  Naturally occurring tritium is
extremely rare on Earth, where trace amounts are formed by the interaction of
the atmosphere with cosmic rays.  It can be produced by irradiating lithium
metal or lithium-bearing ceramic pebbles in a nuclear reactor.

Ref:  https://en.wikipedia.org/wiki/Tritium
''' ),

    'triple_point_of_water' :
        RPNConstantInfo( '273.16', 'kelvin', [ ], True,
                         'the triple point of water',
                         '''
In thermodynamics, the triple point of a substance is the temperature and
pressure at which the three phases (gas, liquid, and solid) of that substance
coexist in thermodynamic equilibrium.

The triple point of water was used to define the kelvin, the base unit of
thermodynamic temperature in the International System of Units (SI).  The value
of the triple point of water was fixed by definition, rather than measured, but
that changed with the 2019 redefinition of SI base units.

Ref:  https://en.wikipedia.org/wiki/Triple_point
''' ),

    # heavenly body constants
    'sun_luminosity' :
        RPNConstantInfo( '3.826e26', 'watt', [ 'solar_luminosity' ], True,
                         'the average luminosity of the Sun in watts',
                         '''
The watt (symbol: W) is a unit of power.  In the International System of Units
(SI) it is defined as a derived unit of 1 joule per second, and is used to
quantify the rate of energy transfer.

Ref:  https://en.wikipedia.org/wiki/Watt
''' ),

    'sun_mass' :
        RPNConstantInfo( '1.988500e30', 'kilogram', [ 'solar_mass' ], True,
                         'the estimated mass of the Sun',
                         '''
This is the estimated mass of the Sun.

Ref:  http://nssdc.gsfc.nasa.gov/planetary/factsheet/sunfact.html
''' ),

    'sun_radius' :
        RPNConstantInfo( '6.9599e8', 'meter', [ 'solar_radius' ], True,
                         'the estimated equitorial radius of the sun',
                         '''
This is the estimated equitorial radius of the Sun.
''' ),

    'sun_volume' :
        RPNConstantInfo( '1.412e27', 'meter^3', [ 'solar_volume' ], True,
                         'the estimated volume of the sun',
                         '''
This is the estimated volume of the Sun.
''' ),


    'mercury_mass' :
        RPNConstantInfo( '3.301e26', 'kilogram', [ ], True,
                         'the estimated mass of the planet Mercury',
                         '''
This is the estimated mass of the planet Mercury.
''' ),

    # equitorial radius
    'mercury_radius' :
        RPNConstantInfo( '2.4397e6', 'meter', [ ], True,
                         'the estimated radius of the planet Mercury',
                         '''
This is the estimated equitorial radius of the planet Mercury.
''' ),

    # sidereal orbit period
    'mercury_revolution' :
        RPNConstantInfo( '87.969', 'day', [ 'mercury_year' ], True,
                         'the observed sidereal orbit period (year) of the planet Mercury',
                         '''
This is the observed sidereal orbit period (year) of the planet Mercury.
''' ),

    'mercury_volume' :
        RPNConstantInfo( '6.083e19', 'meter^3', [ ], True,
                         'the estimated volume of the planet Mercury',
                         '''
This is the estimated volume of the planet Mercury.
''' ),

    'venus_mass' :
        RPNConstantInfo( '4.8689952e24', 'kilogram', [ ], True,
                         'the estimated mass of the planet Venus',
                         '''
This is the estimated mass of the planet Venus.
''' ),

    'venus_radius' :
        RPNConstantInfo( '6.0518e6', 'meter', [ ], True,
                         ' the estimated equitorial radius of the planet Venus',
                         '''
This is the estimated equitorial radius of the planet Venus.
''' ),

    'venus_revolution' :
        RPNConstantInfo( '224.701', 'day', [ 'venus_year' ], True,
                         'the observed sidereal orbit period (year) of the planet Venus',
                         '''
This is the observed sidereal orbit period (year) of the planet Venus.
''' ),

    'venus_volume' :
        RPNConstantInfo( '9.2843e20', 'meter^3', [ ], True,
                         'the estimated volume of the planet Venus',
                         '''
This is the estimated volume of the planet Venus.
''' ),

    'earth_density' :
        # https://en.wikipedia.org/wiki/Earth#Composition_and_structure
        RPNConstantInfo( '5.514', 'gram/centimeter^3', [ ], True,
                         'the estimated average density of the planet Earth',
                         '''
This is the estimated average density of the planet Earth.
''' ),

    'earth_gravity' :
        RPNConstantInfo( '9.80665', 'meter/second^2',               # based on earth_radius
                         [ 'earth_gravities', 'gee', 'gees', 'standard_gravity', 'standard_gravities' ], True,
                         'the average surface gravity of the planet Earth',
                         '''
This is the average surface gravity of the planet Earth.
''' ),

    'earth_mass' :
        RPNConstantInfo( '5.9640955e24', 'kilogram', [ ], True,     # based on earth_radius and earth_gravity
                         'the estimated mass of the planet Earth',
                         '''
This is the estimated mass of the planet Earth.
''' ),

    'earth_radius' :
        # https://en.wikipedia.org/wiki/Earth_radius#Global_average_radii - volumetric radius
        RPNConstantInfo( '6371800', 'meter', [ ], True,
                         'the estimated equitorial radius of the planet Earth',
                         '''
This is the estimated equitorial radius of the planet Earth.
''' ),

    'earth_volume' :
        RPNConstantInfo( '1.083207324897e21', 'meter^3', [ ], True, # based on earth_radius
                         'the estimated volume of the planet Earth',
                         '''
This is the estimated volume of the planet Earth.
''' ),

    'sidereal_year' :
        RPNConstantInfo( '365.256360417', 'day', [ ], True,
                         'the observed sidereal orbit period (year) of the planet Earth',
                         '''
This is the observed sidereal orbit period (year) of the planet Earth.
''' ),

    'tropical_year' :
        RPNConstantInfo( '365.24219', 'day', [ 'solar_year' ], True,
                         'the mean tropical year of the planet Earth',
                         '''
The definition used is the calculation of the mean tropical year on
1 January 2000.
''' ),

    'moon_gravity' :
        RPNConstantInfo( '1.62', 'meter/second^2',                  # based on moon_radius
                         [ 'moon_gravities', 'lunar_gravity', 'lunar_gravities' ], True,
                         'the average surface gravity of the Moon',
                         '''
This is the average surface gravity of the Moon.
''' ),

    'moon_mass' :
        RPNConstantInfo( '7.342e22', 'kilogram', [ 'lunar_mass' ], True,
                         'the estimated mass of the Moon',
                         '''
This is the estimated mass of the Moon.
''' ),


    'moon_radius' :
        RPNConstantInfo( '1.7381e6', 'meter', [ 'lunar_radius' ], True,
                         'the estimated equitorial radius of the Moon',
                         '''
This is the estimated equitorial radius of the Moon.
''' ),

    'moon_revolution' :
        RPNConstantInfo( '27.321662', 'day', [ 'sidereal_month', 'lunar_revolution' ], True,
                         'the observed sidereal orbit period of the Moon around the Earth',
                         '''
This is the observed sidereal orbit period of the Moon around the Earth.
''' ),

    'moon_volume' :
        RPNConstantInfo( '2.1958e19', 'meter^3', [ 'lunar_volume' ], True,
                         'the estimated volume of the Moon',
                         '''
This is the estimated volume of the Moon.
''' ),

    'mars_mass' :
        RPNConstantInfo( '6.4191269e23', 'kilogram', [ ], True,
                         'the estimated mass of the planet Mars',
                         '''
This is the estimated mass of the planet Mars.
''' ),

    'mars_radius' :
        RPNConstantInfo( '3.3962e6', 'meter', [ ], True,
                         'the estimated equitorial radius of the planet Mars',
                         '''
This is the estimated equitorial radius of the planet Mars.
''' ),

    'mars_revolution' :
        RPNConstantInfo( '686.980', 'day', [  'mars_year' ], True,
                         'the observed sidereal orbit period (year) of the planet Mars',
                         '''
This is the observed sidereal orbit period (year) of the planet Mars.
''' ),

    'mars_volume' :
        RPNConstantInfo( '1.6318e20', 'meter^3', [ ], True,
                         'the estimated volume of the planet Mars',
                         '''
This is the estimated volume of the planet Mars.
''' ),

    'jupiter_mass' :
        RPNConstantInfo( '1.8983e27', 'kilogram', [ ], True,
                         'the estimated mass of the planet Jupiter',
                         '''
This is the estimated mass of the planet Jupiter.
''' ),

    'jupiter_radius' :
        RPNConstantInfo( '7.1492e7', 'meter', [ ], True,
                         'the estimated equitorial radius of the planet Jupiter',
                         '''
This is the estimated equitorial radius of the planet Jupiter.
''' ),

    'jupiter_revolution' :
        RPNConstantInfo( '11.862', 'year', [ 'jupiter_year' ], True,
                         'the observed sidereal orbit period (year) of the planet Jupiter',
                         '''
This is the observed sidereal orbit period (year) of the planet Jupiter.
''' ),

    'jupiter_volume' :
        RPNConstantInfo( '1.43128e24', 'meter^3', [ ], True,
                         'the estimated volume of the planet Jupiter',
                         '''
This is the estimated volume of the planet Jupiter.
''' ),

    'saturn_mass' :
        RPNConstantInfo( '5.6836e26', 'kilogram', [ ], True,
                         'the estimated mass of the planet Saturn',
                         '''
This is the estimated mass of the planet Saturn.
''' ),

    'saturn_radius' :
        RPNConstantInfo( '6.0268e7', 'meter', [ ], True,
                         'the estimated equitorial radius of the planet Saturn',
                         '''
This is the estimated equitorial radius of the planet Saturn.
''' ),

    'saturn_revolution' :
        RPNConstantInfo( '29.457', 'year', [ 'saturn_year' ], True,
                         'the observed sidereal orbit period (year) of the planet Saturn',
                         '''
This is the observed sidereal orbit period (year) of the planet Saturn.
''' ),

    'saturn_volume' :
        RPNConstantInfo( '8.2713e23', 'meter^3', [ ], True,
                         'the estimated volume of the planet Saturn',
                         '''
This is the estimated volume of the planet Saturn.
''' ),

    'uranus_mass' :
        RPNConstantInfo( '8.6816e25', 'kilogram', [ ], True,
                         'the estimated mass of the planet Uranus',
                         '''
This is the estimated mass of the planet Uranus.
''' ),

    'uranus_radius' :
        RPNConstantInfo( '2.5559e7', 'meter', [ ], True,
                         'the estimated equitorial radius of the planet Uranus',
                         '''
This is the estimated equitorial radius of the planet Uranus.
''' ),

    'uranus_revolution' :
        RPNConstantInfo( '84.011', 'year', [ 'uranus_year' ], True,
                         'the observed sidereal orbit period (year) of the planet Uranus',
                         '''
This is the observed sidereal orbit period (year) of the planet Uranus.
''' ),

    'uranus_volume' :
        RPNConstantInfo( '6.833e22', 'meter^3', [ ], True,
                         'the estimated volume of the planet Uranus',
                         '''
This is the estimated volume of the planet Uranus.
''' ),

    'neptune_mass' :
        RPNConstantInfo( '1.0242e26', 'kilogram', [ ], True,
                         'the estimated mass of the planet Neptune',
                         '''
This is the estimated mass of the planet Neptune.
''' ),

    'neptune_radius' :
        RPNConstantInfo( '2.4764e7', 'meter', [ ], True,
                         'the estimated equitorial radius of the planet Neptune',
                         '''
This is the estimated equitorial radius of the planet Neptune.
''' ),

    'neptune_revolution' :
        RPNConstantInfo( '164.79', 'year', [ 'neptune_year' ], True,
                         'the observed sidereal orbit period (year) of the planet Neptune',
                         '''
This is the observed sidereal orbit period (year) of the planet Neptune.
''' ),

    'neptune_volume' :
        RPNConstantInfo( '6.254e22', 'meter^3', [ ], True,
                         'the estimated volume of the planet Neptune',
                         '''
This is the estimated volume of the planet Neptune.
''' ),

    'pluto_mass' :
        RPNConstantInfo( '1.0303e22', 'kilogram', [ ], True,
                         'the estimated mass of the planet Pluto',
                         '''
This is the estimated mass of the planet Pluto.

Yes, I still count Pluto as a planet.
''' ),

    'pluto_radius' :
        RPNConstantInfo( '1.185e6', 'meter', [ ], True,
                         'the estimated equitorial radius of the planet Pluto',
                         '''
This is the estimated equitorial radius of the planet Pluto.

Yes, I still count Pluto as a planet.
''' ),

    'pluto_revolution' :
        RPNConstantInfo( '247.94', 'year', [ 'pluto_year' ], True,
                         'the observed sidereal orbit period (year) of the planet Pluto',
                         '''
This is the observed sidereal orbit period (year) of the planet Pluto.

Yes, I still count Pluto as a planet.
''' ),

    'pluto_volume' :
        RPNConstantInfo( '6.97e18', 'meter^3', [ ], True,
                         'the estimated volume of the planet Pluto',
                         '''
This is the estimated volume of the planet Pluto.

Yes, I still count Pluto as a planet.
''' ),

    # settings constants
    'default' :
        RPNConstantInfo( '-1', '', [ ], False,
                         '\'default\' simply evaluates to -1',
                         '''
This is defined for convenience for used with settings operators.
''' ),

    'false' :
        RPNConstantInfo( 0, '', [ ], False,
                         '\'false\' simply evaluates to 0',
                         '''
This is defined for convenience for used with settings operators.
''' ),

    'true' :
        RPNConstantInfo( 1, '', [ ], False,
                         '\'true\' simply evaluates to 1',
                         '''
This is defined for convenience for used with settings operators.

rpn (1)>5 12 **
244140625
rpn (2)>true comma
1
rpn (3)>5 12 **
244,140,625
''' ),

    # day of week constants
    'monday' :
        RPNConstantInfo( 1, '', [ 'mon' ], False,
                         'a constant for the day Monday (1)',
                         '''
This is defined for convenience for use with date operators and has a value of
1.
''' ),

    'tuesday' :
        RPNConstantInfo( 2, '', [ 'tue', 'tues' ], False,
                         'a constant for the day Tuesday (2)',
                         '''
This is defined for convenience for use with date operators and has a value of
2.
''' ),

    'wednesday' :
        RPNConstantInfo( 3, '', [ 'wed' ], False,
                         'a constant for the day Wednesday (3)',
                         '''
This is defined for convenience for use with date operators and has a value of
3.
''' ),

    'thursday' :
        RPNConstantInfo( 4, '', [ 'thu', 'thur','thurs' ], False,
                         'a constant for the day Thursday (4)',
                         '''
This is defined for convenience for use with date operators and has a value of
4.
''' ),

    'friday' :
        RPNConstantInfo( 5, '', [ 'fri' ], False,
                         'a constant for the day Friday (5)',
                         '''
This is defined for convenience for use with date operators and has a value of
5.
''' ),

    'saturday' :
        RPNConstantInfo( 6, '', [ 'sat' ], False,
                         'a constant for the day Saturday (6)',
                         '''
This is defined for convenience for use with date operators and has a value of
6.
''' ),

    'sunday' :
        RPNConstantInfo( 7, '', [ ], False,
                         'a constant for the day Sunday (7)',
                         '''
This is defined for convenience for use with date operators and has a value of
7.
''' ),

    # month constants
    'january' :
        RPNConstantInfo( 1, '', [ 'jan' ], False,
                         'a constant for the month of January (1)',
                         '''
This is defined for convenience for use with date operators and has a value of
1.
''' ),

    'february' :
        RPNConstantInfo( 2, '', [ 'feb' ],  False,
                         'a constant for the month of February (2)',
                         '''
This is defined for convenience for use with date operators and has a value of
2.
''' ),

    'march' :
        RPNConstantInfo( 3, '', [ 'mar' ], False,
                         'a constant for the month of March (3)',
                         '''
This is defined for convenience for use with date operators and has a value of
3.
''' ),

    'april' :
        RPNConstantInfo( 4, '', [ 'apr' ], False,
                         'a constant for the month of April (4)',
                         '''
This is defined for convenience for use with date operators and has a value of
4.
''' ),

    'may' :
        RPNConstantInfo( 5, '', [ ], False,
                         'a constant for the month of May (5)',
                         '''
This is defined for convenience for use with date operators and has a value of
5.
''' ),

    'june' :
        RPNConstantInfo( 6, '', [ 'jun' ], False,
                         'a constant for the month of June (6)',
                         '''
This is defined for convenience for use with date operators and has a value of
6.
''' ),

    'july' :
        RPNConstantInfo( 7, '', [ 'jul' ], False,
                         'a constant for the month of July (7)',
                         '''
This is defined for convenience for use with date operators and has a value of
7.
''' ),

    'august' :
        RPNConstantInfo( 8, '', [ 'aug' ], False,
                         'a constant for the month of August (8)',
                         '''
This is defined for convenience for use with date operators and has a value of
8.
''' ),

    'september' :
        RPNConstantInfo( 9, '', [ 'sep' ], False,
                         'a constant for the month of September (9)',
                         '''
This is defined for convenience for use with date operators and has a value of
9.
''' ),

    'october' :
        RPNConstantInfo( 10, '', [ 'oct' ], False,
                         'a constant for the month of October (10)',
                         '''
This is defined for convenience for use with date operators and has a value of
10.
''' ),

    'november' :
        RPNConstantInfo( 11, '', [ 'nov' ], False,
                         'a constant for the month of November (11)',
                         '''
This is defined for convenience for use with date operators and has a value of
11.
''' ),

    'december' :
        RPNConstantInfo( 12, '', [ 'dec' ], False,
                         'a constant for the month of December (12)',
                         '''
This is defined for convenience for use with date operators and has a value of
12.
''' ),

    # programming integer constants
    'max_char' :
        RPNConstantInfo( ( 1 << 7 ) - 1, '', [ 'maxchar', 'max_int8', 'maxint8' ], False,
                         'the maximum 8-bit signed integer',
                         '''
This is the largest number that can be represented by an 8-bit signed
integer assuming two's complement representation.
''' ),

    'max_double' :
        RPNConstantInfo( getMaxDouble( ), '', [ 'maxdouble' ], False,
                         'the largest value that can be represented by a 64-bit IEEE 754 float',
                         '''
For all IEEE 754 floating point numbers, rpn assumes big-endian byte ordering.
''' ),

    'max_float' :
        RPNConstantInfo( getMaxFloat( ), '', [ 'maxfloat' ], False,
                         'the largest value that can be represented by a 32-bit IEEE 754 float',
                         '''
For all IEEE 754 floating point numbers, rpn assumes big-endian byte ordering.
''' ),

    'max_long' :
        RPNConstantInfo( ( 1 << 31 ) - 1, '',
                         [ 'max_int', 'maxint', 'max_int32', 'maxint32', 'maxlong' ], False,
                         'the maximum 32-bit signed integer',
                         '''
This is the largest number that can be represented by a 32-bit signed
integer assuming two's complement representation.
''' ),

    'max_longlong' :
        RPNConstantInfo( ( 1 << 63 ) - 1, '', [ 'max_int64', 'maxint64', 'maxlonglong' ], False,
                         'the maximum 64-bit signed integer',
                         '''
This is the largest number that can be represented by a 64-bit signed
integer assuming two's complement representation.
''' ),

    'max_quadlong' :
        RPNConstantInfo( ( 1 << 127 ) - 1, '',
                         [ 'max_int128', 'maxint128', 'max_quad', 'maxquad', 'maxquadlong' ], False,
                         'the maximum 128-bit signed integer',
                         '''
This is the largest number that can be represented by a 128-bit signed
integer assuming two's complement representation.
''' ),

    'max_short' :
        RPNConstantInfo( ( 1 << 15 ) - 1, '', [ 'max_int16', 'maxint16', 'maxshort' ], False,
                         'the maximum 16-bit signed integer',
                         '''
This is the largest number that can be represented by a 16-bit signed
integer assuming two's complement representation.
''' ),

    'max_uchar' :
        RPNConstantInfo( ( 1 << 8 ) - 1, '', [ 'max_uint8', 'maxuint8', 'maxuchar' ], False,
                         'the maximum 8-bit unsigned integer',
                         '''
This is the largest number that can be represented by an 8-bit unsigned
integer assuming two's complement representation.
''' ),

    'max_ulong' :
        RPNConstantInfo( ( 1 << 32 ) - 1, '',
                         [ 'max_uint32', 'maxuint32', 'max_uint', 'maxuint', 'maxulong' ], False,
                         'the maximum 32-bit unsigned integer',
                         '''
This is the largest number that can be represented by a 32-bit unsigned
integer assuming two's complement representation.
''' ),

    'max_ulonglong' :
        RPNConstantInfo( ( 1 << 64 ) - 1, '', [ 'max_uint64', 'maxuint64', 'maxulonglong' ], False,
                         'the maximum 64-bit unsigned integer',
                         '''
This is the largest number that can be represented by a 64-bit unsigned
integer assuming two's complement representation.
''' ),

    'max_uquadlong' :
        RPNConstantInfo( ( 1 << 128 ) - 1, '', [ 'max_uint128', 'maxuint128', 'maxuquadlong' ], False,
                         'the maximum 128-bit unsigned integer',
                         '''
This is the largest number that can be represented by a 128-bit unsigned
integer assuming two's complement representation.
''' ),

    'max_ushort' :
        RPNConstantInfo( ( 1 << 16 ) - 1, '', [ 'max_uint16', 'maxuint16', 'maxushort' ], False,
                         'the maximum 16-bit unsigned integer',
                         '''
This is the largest number that can be represented by a 16-bit unsigned
integer assuming two's complement representation.
''' ),

    'min_char' :
        RPNConstantInfo( -( 1 << 7 ), '', [ 'min_int8', 'minint8', 'minchar' ], False,
                         'the minimum 8-bit signed integer',
                         '''
This is the smallest number that can be represented by an 8-bit signed
integer assuming two's complement representation.
''' ),

    'min_double' :
        RPNConstantInfo( getMinDouble( ), '', [ 'mindouble' ], False,
                         'the smallest value that can be represented by a 64-bit IEEE 754 float',
                         '''
For all IEEE 754 floating point numbers, rpn assumes big-endian byte ordering.
''' ),

    'min_float' :
        RPNConstantInfo( getMinFloat( ), '', [ 'minfloat' ], False,
                         'the smallest value that can be represented by a 32-bit IEEE 754 float',
                         '''
For all IEEE 754 floating point numbers, rpn assumes big-endian byte ordering.
''' ),

    'min_long' :
        RPNConstantInfo( -( 1 << 31 ), '', [ 'min_int32', 'minint32', 'minlong' ], False,
                         'the minimum 32-bit signed integer',
                         '''
This is the smallest number that can be represented by a 32-bit signed
integer assuming two's complement representation.
''' ),

    'min_longlong' :
        RPNConstantInfo( -( 1 << 63 ), '', [ 'min_int64', 'minint64', 'minlonglong' ], False,
                         'the minimum 64-bit signed integer',
                         '''
This is the smallest number that can be represented by a 64-bit signed
integer assuming two's complement representation.
''' ),

    'min_quadlong' :
        RPNConstantInfo( -( 1 << 127 ), '',
                         [ 'min_int128', 'minint128', 'min_quad', 'minquad', 'minquadlong' ], False,
                         'the minimum 128-bit signed integer',
                         '''
This is the smallest number that can be represented by a 128-bit signed
integer assuming two's complement representation.
''' ),

    'min_short' :
        RPNConstantInfo( -( 1 << 15 ), '', [ 'min_int16', 'minint16', 'minshort' ], False,
                         'the minimum 16-bit signed integer',
                         '''
This is the smallest number that can be represented by a 16-bit signed
integer assuming two's complement representation.
''' ),

    'min_uchar' :
        RPNConstantInfo( 0, '', [ 'min_uint8', 'minuint8', 'minuchar' ], False,
                         'the minimum 8-bit unsigned integer',
                         '''
This is the smallest number that can be represented by an 8-bit unsigned
integer assuming two's complement representation.
''' ),

    'min_ulong' :
        RPNConstantInfo( 0, '', [ 'min_uint32', 'minuint32', 'minulong' ], False,
                         'the minimum 32-bit unsigned integer',
                         '''
This is the smallest number that can be represented by a 32-bit unsigned
integer assuming two's complement representation.
''' ),

    'min_ulonglong' :
        RPNConstantInfo( 0, '', [ 'min_uint64', 'minuint64', 'minulonglong' ], False,
                         'the minimum 64-bit unsigned integer',
                         '''
This is the smallest number that can be represented by a 64-bit unsigned
integer assuming two's complement representation.
''' ),

    'min_uquadlong' :
        RPNConstantInfo( 0, '', [ 'min_uint128', 'minuint128', 'min_quad', 'minquad', 'minquadlong' ], False,
                         'the minimum 128-bit unsigned integer',
                         '''
This is the smallest number that can be represented by a 128-bit unsigned
integer assuming two's complement representation.
''' ),

    'min_ushort' :
        RPNConstantInfo( 0, '', [ 'min_uint16', 'minuint16', 'minushort' ], False,
                         'the minimum 16-bit unsigned integer',
                         '''
This is the smallest number that can be represented by a 16-bit unsigned
integer assuming two's complement representation.
''' ),

    # mathematical constants
    'i' :
        RPNConstantInfo( 1j, '', [ ], False,
                         'i, the square root of -1, which is the same as \'1j\'',
                         '''
The imaginary number i is the square root of -1.  Python's normal syntax for
imaginary numbers uses the 'j' suffix with a number, but for just the value of i
itself, '1j' seems clumsy, so there's the 'i' constant.

'-i' however cannot be used since that is the command-line switch for
interactive mode.
''' ),

}




#!/usr/bin/env python

#******************************************************************************
#
#  rpn.py
#
#  rpnChilada - RPN command-line calculator
#  copyright (c) 2020, Rick Gutleber (rickg@his.com)
#
#  *** NOTE:  Don't run this file directly.  Use ../rpn.py.
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

#  https://en.wikipedia.org/wiki/Medieval_weights_and_measures

#  http://mathworld.wolfram.com/FavardConstants.html
#  How can we create a lambda to produce multiple Favard constants with a single expression?

#  rpn 4 meters meters * is_square

import argparse
import os
import sys
import time

from pathlib import Path

try:
    import readline
except ImportError:
    import pyreadline as readline

from mpmath import fneg, im, mp, mpc, mpmathify, nan, nstr, re

from rpn.rpnAliases import operatorAliases
from rpn.rpnBase import specialBaseNames
from rpn.rpnConstantUtils import loadGlobalConstants
from rpn.rpnDateTime import RPNDateTime
from rpn.rpnDebug import debugPrint
from rpn.rpnGenerator import RPNGenerator
from rpn.rpnMeasurement import RPNMeasurement

from rpn.rpnOperator import checkForVariable

from rpn.rpnOperators import evaluateTerm, functionOperators, loadUserFunctionsFile, \
                             RPNFunction, saveUserFunctionsFile, setAccuracy, setPrecision

from rpn.rpnOutput import formatDateTime, formatListOutput, formatOutput, formatUnits, \
                          printHelp, printHelpModeHelp, printInteractiveHelp, printTitleScreen

from rpn.rpnPersistence import loadUnitData, loadUnitNameData, loadUserVariablesFile, \
                               saveUserVariablesFile, loadUserConfigurationFile, saveResult, \
                               saveUserConfigurationFile

from rpn.rpnPrimeUtils import checkForPrimeData

from rpn.rpnSpecial import handleIdentify

from rpn.rpnUtils import getCurrentArgList, getUserDataPath, parseNumerals, validateArguments, \
                         validateOptions

from rpn.rpnVersion import RPN_PROGRAM_NAME, PROGRAM_NAME, PROGRAM_VERSION, PROGRAM_DESCRIPTION, \
                           COPYRIGHT_MESSAGE

import rpn.rpnGlobals as g

if not hasattr( time, 'time_ns' ):
    from rpn.rpnNanoseconds import time_ns
else:
    from time import time_ns




#******************************************************************************
#
#  lookAhead
#
#******************************************************************************

def lookAhead( iterable ):
    '''
    Pass through all values from the given iterable, augmented by the
    information if there are more values to come after the current one
    (True), or if it is the last value (False).
    '''
    # Get an iterator and pull the first value.
    i = iter( iterable )
    last = next( i )

    # Run the iterator to exhaustion (starting from the second value).
    for value in i:
        # Report the *previous* value (more to come).
        yield last, True
        last = value

    # Report the last value.
    yield last, False


#******************************************************************************
#
#  evaluate
#
#******************************************************************************

def evaluate( terms ):
    '''
    This is the core function in rpn that evaluates the terms to be
    calculated.  terms are put into a stack, and popped off one at a time
    when evaluated.  When an operator is popped off and evaluated, the results
    get pushed back on to the stack for further processing, or ultimately,
    output.
    '''
    valueList = list( )
    index = 1                 # only used for error messages

    # handle a unit operator
    if not g.unitOperators:
        loadUnitData( )
        loadGlobalConstants( )

    # start parsing terms and populating the evaluation stack... this is the heart of rpn
    for term, hasMore in lookAhead( terms ):
        if term in g.aliases:
            term = g.aliases[ term ]

        if term in functionOperators:
            if g.creatingFunction:
                g.creatingFunction = False
            else:
                raise ValueError( 'function operators require a function definition' )

        currentValueList = getCurrentArgList( valueList )

        if g.creatingFunction:
            currentValueList[ -1 ].add( term )
            continue

        try:
            if not evaluateTerm( term, index, currentValueList, not hasMore ):
                valueList = [ nan ]
                break
        except ValueError as error:
            print( 'rpn:  error:  {0}'.format( error ) )
            valueList = [ nan ]
            break

        index = index + 1

    if len( g.echoArguments ) > 0:
        returnValue = list( g.echoArguments )

        if isinstance( valueList, list ) and len( valueList ) == 1:
            returnValue.append( valueList[ 0 ] )
        else:
            returnValue.append( valueList )

        g.echoArguments = [ ]
        return [ returnValue ]

    return valueList


#******************************************************************************
#
#  handleOutput
#
#******************************************************************************

def handleOutput( valueList, indent=0, file=sys.stdout ):
    '''
    Once the evaluation of terms is complete, the results need to be
    translated into output.

    If the result is a list or a generator, special formatting turns those
    into text output.  Date-time values and measurements also require special
    formatting.

    Setting file to an io.StringIO objects allows for 'printing' to a string,
    which is used by makeHelp.py to generate actual rpn output for the examples.
    '''
    if valueList is None:
        return file

    indentString = ' ' * indent

    if len( valueList ) != 1:
        if g.checkForSingleResults:
            print( 'valueList', valueList )
            raise ValueError( 'unexpected multiple results!' )

        valueList = [ valueList ]

    if isinstance( valueList[ 0 ], RPNFunction ):
        print( indentString + 'rpn:  unexpected end of input in function definition', file=file )
    else:
        mp.pretty = True
        result = valueList.pop( )

        if result is nan:
            return file

        if g.comma:
            g.integerGrouping = 3     # override whatever was set on the command-line
            g.leadingZero = False     # this one, too
            g.integerDelimiter = ','
        else:
            g.integerDelimiter = ' '

        if isinstance( result, RPNGenerator ):
            formatListOutput( result.getGenerator( ), indent=indent, file=file )
        elif isinstance( result, list ):
            formatListOutput( result, indent=indent, file=file )
        else:
            # single result
            if isinstance( result, RPNDateTime ):
                outputString = formatDateTime( result )
            elif isinstance( result, str ):
                result = checkForVariable( result )
                outputString = result
            else:
                # output the answer with all the extras according to command-line arguments
                # handle the units if we are displaying a measurement
                if isinstance( result, RPNMeasurement ):
                    outputString = formatOutput( nstr( result.value, g.outputAccuracy, min_fixed=-g.maximumFixed - 1 ) )
                    outputString += ' ' + formatUnits( result )
                # handle a complex output (mpmath type: mpc)
                elif isinstance( result, mpc ):
                    #print( 'result', result, type( result ) )
                    #print( 'im', im( result ), type( im( result ) ) )
                    #print( 're', re( result ), type( re( result ) ) )

                    if im( result ) > 0:
                        outputString = '(' + formatOutput( nstr( mpmathify( re( result ) ), \
                                                           g.outputAccuracy, min_fixed=-g.maximumFixed - 1 ) ) + \
                                       ' + ' + formatOutput( nstr( mpmathify( im( result ) ), \
                                                             g.outputAccuracy, min_fixed=-g.maximumFixed - 1 ) ) + 'j)'
                    elif im( result ) < 0:
                        outputString = '(' + formatOutput( nstr( mpmathify( re( result ) ), \
                                                                 g.outputAccuracy, min_fixed=-g.maximumFixed - 1 ) ) + \
                                       ' - ' + formatOutput( nstr( fneg( mpmathify( im( result ) ) ), \
                                                             g.outputAccuracy, min_fixed=-g.maximumFixed - 1 ) ) + 'i)'
                    else:
                        outputString = formatOutput( nstr( re( result ), g.outputAccuracy,
                                                           min_fixed=-g.maximumFixed - 1 ) )
                # otherwise, it's a plain old mpf
                else:
                    outputString = formatOutput( nstr( result, g.outputAccuracy, min_fixed=-g.maximumFixed - 1 ) )

            print( indentString + outputString, file=file )

            # handle --identify
            if g.identify:
                handleIdentify( result, file )

        saveResult( result )

    if g.timer or g.tempTimerMode:
        print( '\n' + indentString + '{:.3f} seconds'.format( ( time_ns( ) - g.startTime ) / 1000000000 ), file=file )

    return file


#******************************************************************************
#
#  enterInteractiveMode
#
#******************************************************************************

def enterInteractiveMode( ):
    '''
    If rpn is launched with no expression, then it goes into interactive
    mode, where it will continue to evaluate new expressions input until
    the 'exit' command.
    '''
    readline.parse_and_bind( 'tab: complete' )
    readline.parse_and_bind( 'set editing-mode vi' )

    printTitleScreen( PROGRAM_NAME, PROGRAM_DESCRIPTION )

    g.results.append( None )   # g.results[ 0 ]
    g.interactive = True

    while True:
        g.promptCount += 1

        # clear single operation flags
        g.tempCommaMode = False
        g.tempHexMode = False
        g.tempIdentifyMode = False
        g.tempLeadingZeroMode = False
        g.tempOctalMode = False
        g.tempTimerMode = False

        try:
            line = input( 'rpn (' + str( g.promptCount ) + ')> ' )
        except EOFError:
            break

        line = line.strip( )

        if line in [ 'exit', 'quit' ]:
            break

        terms = line.split( ' ' )

        if terms[ 0 ] == 'help':
            enterHelpMode( terms[ 1 : ] )
        else:
            if g.timer or g.tempTimerMode:
                g.startTime = time_ns( )

            #newTerms = preprocessTerms( terms )
            #print( 'newTerms', newTerms )

            if validateArguments( terms ):
                valueList = evaluate( terms )

                g.results.append( valueList[ -1 ] )

                handleOutput( valueList )
            else:
                g.results.append( 0 )


#******************************************************************************
#
#  enterHelpMode
#
#******************************************************************************

def enterHelpMode( terms ):
    '''
    When using rpn interactively, help is a special mode, which allows the user
    to navigate the help contents with much fewer keystrokes than having to
    invoke help over and over.
    '''
    printHelpModeHelp( )

    while True:
        try:
            line = input( 'rpn help>' )
        except EOFError:
            break

        line = line.strip( )

        if line in [ 'exit', 'quit' ]:
            break

        terms = line.split( ' ' )

        if terms[ 0 ] == 'help':
            printHelpModeHelp( )
        elif terms[ 0 ] == 'topics':
            printInteractiveHelp( )
        else:
            for term in terms:
                printHelp( [ term ], interactive=True )


#******************************************************************************
#
#  rpn
#
#******************************************************************************

def rpn( cmdArgs ):
    '''
    This is the main function which processes the command-line arguments,
    handling both options and the expression to evaluate.   This function is
    mainly concerned with parsing and handling the command-line options.

    It finally calls evaluate( ) with the expression to be calculated, and
    returns the results, which can be formatted for output or used in another
    way (such as the unit test functionality).
    '''
    # initialize globals
    g.outputRadix = 10

    # look for help argument before we start setting everything up (because it's faster this way)
    showHelp = False
    helpArgs = [ ]

    for arg in cmdArgs:
        if arg == 'help':
            showHelp = True
        else:
            if showHelp:
                helpArgs.append( arg )

    if showHelp:
        parser = argparse.ArgumentParser( prog = PROGRAM_NAME, description = RPN_PROGRAM_NAME +
                                          ' - ' + PROGRAM_DESCRIPTION + '\n    ' +
                                          COPYRIGHT_MESSAGE, add_help = False,
                                          formatter_class = argparse.RawTextHelpFormatter,
                                          prefix_chars = '-' )

        parser.add_argument( 'terms', nargs = '*', metavar = 'term' )
        parser.add_argument( '-l', '--line_length', type = int, action = 'store',
                             default = g.defaultLineLength )

        args = parser.parse_args( cmdArgs )

        loadUnitNameData( )

        g.aliases.update( operatorAliases )

        printHelp( helpArgs )
        return None

    # set up the command-line options parser
    parser = argparse.ArgumentParser( prog = PROGRAM_NAME, description = RPN_PROGRAM_NAME +
                                      ' - ' + PROGRAM_DESCRIPTION + '\n    ' +
                                      COPYRIGHT_MESSAGE, add_help = False,
                                      formatter_class = argparse.RawTextHelpFormatter,
                                      prefix_chars = '-' )

    parser.add_argument( '-a', '--output_accuracy', nargs='?', type=int, default=g.defaultOutputAccuracy,
                         const=g.defaultOutputAccuracy )
    parser.add_argument( '-b', '--input_radix', type=str, default=g.defaultInputRadix )
    parser.add_argument( '-c', '--comma', action='store_true' )
    parser.add_argument( '-d', '--decimal_grouping', nargs='?', type=int, default=0,
                         const=g.defaultDecimalGrouping )
    parser.add_argument( '-D', '--DEBUG', action='store_true' )
    parser.add_argument( '-e', '--profile', action='store_true' )
    parser.add_argument( '-E', '--echo_command', action='store_true' )
    parser.add_argument( '-g', '--integer_grouping', nargs='?', type=int, default=0,
                         const=g.defaultIntegerGrouping )
    parser.add_argument( '-h', '--help', action='store_true' )
    parser.add_argument( '-i', '--identify', action='store_true' )
    parser.add_argument( '-I', '--ignore_cache', action='store_true' )
    parser.add_argument( '-l', '--line_length', type=int, default=g.defaultLineLength )
    parser.add_argument( '-m', '--maximum_fixed', type=int, default=g.defaultMaximumFixed )
    parser.add_argument( '-n', '--numerals', type=str, default=g.defaultNumerals )
    parser.add_argument( '-o', '--octal', action='store_true' )
    parser.add_argument( '-p', '--precision', type=int, default=g.defaultPrecision )
    parser.add_argument( '-r', '--output_radix', type=str, default=g.defaultOutputRadix )
    parser.add_argument( '-s', '--list_format_level', nargs='?', type=int, default=0,
                         const=g.defaultListFormatLevel )
    parser.add_argument( '-t', '--timer', action='store_true' )
    parser.add_argument( '-T', '--time_limit', nargs='?', type=int, default=0, const=g.timeLimit )
    parser.add_argument( '-V', '--version', action='store_true' )
    parser.add_argument( '-v', '--verbose', action='store_true' )
    parser.add_argument( '-w', '--bitwise_group_size', type = int, default = g.defaultBitwiseGroupSize )
    parser.add_argument( '-x', '--hex', action='store_true' )
    parser.add_argument( '-z', '--leading_zero', action='store_true' )
    parser.add_argument( '-!', '--print_options', action='store_true' )
    parser.add_argument( '-?', '--other_help', action='store_true' )

    # pull out the options and the terms
    options = [ ]
    terms = [ ]

    loadUserVariablesFile( )
    loadUserFunctionsFile( )
    loadUserConfigurationFile( )

    if 'yafu_path' in g.userConfiguration and 'yafu_binary' in g.userConfiguration:
        g.useYAFU = True
    else:
        g.useYAFU = False

    for arg in cmdArgs:
        if len( arg ) > 1 :
            if arg[ 0 ] == '$' and arg[ 1 : ] not in g.userVariables:
                raise ValueError( 'undefined user variable referenced: ' + arg )

            if arg[ 0 ] == '@' and arg[ 1 : ] not in g.userFunctions:
                raise ValueError( 'undefined user function referenced: ' + arg )

            if arg[ 0 ] == '-' :
                if arg[ 1 ].isdigit( ):     # a negative number, not an option
                    terms.append( arg )
                else:
                    options.append( arg )
            else:
                terms.append( arg )
        else:
            terms.append( arg )

    debugPrint( 'terms', terms )
    debugPrint( 'options', options )

    # OK, let's parse and validate the options
    args = parser.parse_args( options )

    g.aliases.update( operatorAliases )

    if args.help or args.other_help:
        loadUnitNameData( )

        printHelp( )
        return None

    valid, errorString = validateOptions( args )

    if not valid:
        print( 'rpn:  ' + errorString )
        return None

    # these are either globals or can be modified by other options (like -x)
    g.bitwiseGroupSize = args.bitwise_group_size
    g.integerGrouping = args.integer_grouping
    g.leadingZero = args.leading_zero

    # handle -a
    setAccuracy( args.output_accuracy )

    # handle -b
    g.inputRadix = int( args.input_radix )

    # handle -c
    g.comma = args.comma

    # handle -d
    g.decimalGrouping = args.decimal_grouping

    # handle -D
    if args.DEBUG:
        g.debugMode = True

    # handle -e
    g.echo_command = args.echo_command

    # handle -g
    g.integerGrouping = args.integer_grouping

    # handle -i
    g.identify = args.identify

    # handle -I
    g.ignoreCache = args.ignore_cache

    # handle -l
    g.lineLength = args.line_length

    # handle -m
    g.maximumFixed = args.maximum_fixed

    # handle -n
    g.numerals = parseNumerals( args.numerals )

    # handle -o
    if args.octal:
        g.outputRadix = 8
        g.leadingZero = True
        g.integerGrouping = 3
        g.bitwiseGroupSize = 9

    # handle -p
    setPrecision( args.precision )

    # handle -r
    if args.output_radix in specialBaseNames:
        g.outputRadix = specialBaseNames[ args.output_radix ]
    else:
        try:
            # if g.outputRadix was already set (e.g., by -o) then we don't want to override it
            if g.outputRadix == 10:
                g.outputRadix = int( args.output_radix )
        except ValueError:
            print( 'rpn:  cannot interpret output radix \'%s\' as a number' % args.output_radix )
            return [ nan ]

    # -r validation
    if ( ( g.outputRadix < g.maxSpecialBase ) or ( g.outputRadix == 0 ) or
         ( g.outputRadix == 1 ) or ( g.outputRadix > 62 ) ):
        print( 'rpn:  output radix must be from 2 to 62, fib, phi, fac, doublefac, square, lucas' )
        return [ nan ]

    # handle -s
    g.listFormatLevel = args.list_format_level

    # handle -t
    g.timer = args.timer

    # handle -T
    g.timeLimit = args.time_limit

    # handle -v
    g.verbose = args.verbose

    # handle -V
    if args.version:
        return [ int( i ) for i in PROGRAM_VERSION.split( '.' ) ]

    # handle -x
    if args.hex:
        g.outputRadix = 16
        g.leadingZero = True
        g.integerGrouping = 4
        g.bitwiseGroupSize = 16

    # handle -u and -y:  mpmath wants precision of at least 53 for these functions
    if args.identify and mp.dps < 53:
        setAccuracy( 53 )

    if args.print_options:
        print( '--output_accuracy:  %d' % g.outputAccuracy )
        print( '--input_radix:  %d' % g.inputRadix )
        print( '--comma:  ' + ( 'true' if g.comma else 'false' ) )
        print( '--decimal_grouping:  %d' % g.decimalGrouping )
        print( '--integer_grouping:  %d' % g.integerGrouping )
        print( '--line_length:  %d' % g.lineLength )
        print( '--numerals:  ' + g.numerals )
        print( '--octal:  ' + ( 'true' if args.octal else 'false' ) )
        print( '--precision:  %d' % args.precision )
        print( '--output_radix:  %d' % g.outputRadix )
        print( '--list_format_level:  %d' % g.listFormatLevel )
        print( '--timer:  ' + ( 'true' if args.timer else 'false' ) )
        print( '--verbose:  ' + ( 'true' if g.verbose else 'false' ) )
        print( '--bitwise_group_size:  %d' % g.bitwiseGroupSize )
        print( '--hex:  ' + ( 'true' if args.hex else 'false' ) )
        print( '--identify:  ' + ( 'true' if args.identify else 'false' ) )
        print( '--leading_zero:  ' + ( 'true' if g.leadingZero else 'false' ) )
        print( '--ignore_cache:  ' + ( 'true' if g.ignoreCache else 'false' ) )
        print( )

    g.creatingFunction = False

    # enter interactive mode if there are no arguments
    if not terms:
        if not loadUnitNameData( ):
            return None

        enterInteractiveMode( )
        return None

    # let's check out the arguments before we start to do any calculations
    if not validateArguments( terms ):
        return None

    #newTerms = preprocessTerms( terms )
    #print( 'newTerms', newTerms )

    # waiting until we've validated the arguments to do this because it's slow
    if not loadUnitNameData( ):
        return None

    if g.echo_command:
        print( *sys.argv )

    if g.timer:
        g.startTime = time_ns( )

    return evaluate( terms )


#******************************************************************************
#
#  __main__
#
#******************************************************************************

def main( ):
    checkForPrimeData( )

    unitsFile = Path( getUserDataPath( ) + os.sep + 'units.pckl.bz2' )

    os.chdir( getUserDataPath( ) )     # SkyField doesn't like running in the root directory

    if not unitsFile.is_file( ):
        print( 'Please run "rpnMakeUnits" (or makeUnits.py) to initialize the unit conversion data files.' )
        sys.exit( 0 )

    helpFile = Path( getUserDataPath( ) + os.sep + 'help.pckl.bz2' )

    if not helpFile.is_file( ):
        print( 'Please run "makeHelp" to initialize the help files.' )
        sys.exit( 0 )

    try:
        #for arg in sys.argv:
        #    if arg == '-e':
        #        import profile
        #        profile.run( 'handleOutput( rpn( sys.argv[ 1 : ] ) )' )

        handleOutput( rpn( sys.argv[ 1 : ] ) )

        if g.userVariablesAreDirty:
            saveUserVariablesFile( )

        if g.userFunctionsAreDirty:
            saveUserFunctionsFile( )

        if g.userConfigurationIsDirty:
            saveUserConfigurationFile( )
    except ValueError as error:
        print( '\nrpn:  value error:  {0}'.format( error ) )

        if g.debugMode:
            raise

        sys.exit( 0 )
    except KeyboardInterrupt:
        print( )
        print( 'handling ctrl-c keyboard interrupt...' )
        sys.exit( 0 )


#******************************************************************************
#
#  __main__
#
#******************************************************************************

if __name__ == '__main__':
    main( )




#!/usr/bin/env python

#******************************************************************************
#
#  rpnOutput.py
#
#  rpnChilada output functions
#  copyright (c) 2020, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

import math
import string
import sys
import textwrap

from mpmath import e, floor, frac, im, mp, mpf, mpmathify, nstr, phi, pi, re, sqrt

from rpn.rpnBase import convertFractionToBaseN, convertToBaseN, convertToFibBase, \
                        convertToNonintegerBase, convertToSpecialBase, specialBaseFunctions

from rpn.rpnDateTime import RPNDateTime
from rpn.rpnGenerator import RPNGenerator
from rpn.rpnMeasurement import RPNMeasurement
from rpn.rpnPersistence import loadHelpData, loadUnitData
from rpn.rpnUnitTypes import basicUnitTypes
from rpn.rpnUtils import addAliases
from rpn.rpnVersion import COPYRIGHT_MESSAGE, PROGRAM_VERSION, PROGRAM_VERSION_STRING, \
                           RPN_PROGRAM_NAME, PROGRAM_DESCRIPTION

import rpn.rpnGlobals as g


#******************************************************************************
#
#  formatNumber
#
#  This takes a mpf value and turns it into a string.
#
#******************************************************************************

def formatNumber( number, outputRadix, leadingZero, integerGrouping ):
    negative = ( number < 0 )

    if outputRadix == g.fibBase:
        strInteger = convertToFibBase( floor( number ) )
        strMantissa = ''
    elif outputRadix == g.phiBase:
        strInteger, strMantissa = convertToNonintegerBase( number, phi )
    elif outputRadix == g.eBase:
        strInteger, strMantissa = convertToNonintegerBase( number, e )
    elif outputRadix == g.piBase:
        strInteger, strMantissa = convertToNonintegerBase( number, pi )
    elif outputRadix == g.sqrt2Base:
        strInteger, strMantissa = convertToNonintegerBase( number, sqrt( 2 ) )
    elif outputRadix < 0:   # these mean special bases, not negative numbers
        strInteger = convertToSpecialBase( floor( number ), specialBaseFunctions[ outputRadix ] )
        strMantissa = ''
    elif ( outputRadix != 10 ) or ( g.numerals != g.defaultNumerals ):
        strInteger = str( convertToBaseN( floor( number ), outputRadix, False, g.numerals ) )
        strMantissa = str( convertFractionToBaseN( frac( number ), outputRadix,
                                                   int( mp.dps / math.log10( outputRadix ) ), False ) )
        if strMantissa == '[]':
            strMantissa = ''
    else:
        strNumber = nstr( number, n = g.outputAccuracy, min_fixed=-g.maximumFixed - 1 )

        if '.' in strNumber:
            decimal = strNumber.find( '.' )
        else:
            decimal = len( strNumber )

        strInteger = strNumber[ 1 if negative else 0 : decimal ]

        strMantissa = strNumber[ decimal + 1 : ]

        if strMantissa == '0':
            strMantissa = ''
        elif ( strMantissa != '' ) and ( g.outputAccuracy == -1 ):
            strMantissa = strMantissa.rstrip( '0' )

    if integerGrouping > 0:
        if strInteger[ 0 ] == '-':
            strInteger = strInteger[ 1 : ]

        firstDelimiter = len( strInteger ) % integerGrouping

        if leadingZero and firstDelimiter > 0:
            integerResult = '0' * ( integerGrouping - firstDelimiter )
        else:
            integerResult = ''

        integerResult += strInteger[ : firstDelimiter ]

        for i in range( firstDelimiter, len( strInteger ), integerGrouping ):
            if integerResult != '':
                integerResult += g.integerDelimiter

            integerResult += strInteger[ i : i + integerGrouping ]
    else:
        integerResult = strInteger

    if g.decimalGrouping > 0:
        mantissaResult = ''

        for i in range( 0, len( strMantissa ), g.decimalGrouping ):
            if mantissaResult != '':
                mantissaResult += g.decimalDelimiter

            mantissaResult += strMantissa[ i : i + g.decimalGrouping ]
    else:
        mantissaResult = strMantissa

    result = integerResult

    if mantissaResult != '':
        result += '.' + mantissaResult

    return result, negative


#******************************************************************************
#
#  formatOutput
#
#  This takes a string representation of the result and formats it according
#  to a whole bunch of options.
#
#******************************************************************************

def formatOutput( output ):
    # filter out text strings
    for c in output:
        if c in '+-.':
            continue

        # anything with embedded whitespace is a string
        if c in string.whitespace or c in string.punctuation:
            return output

    #print( )
    #print( 'formatOutput 1:', output )

    # override settings with temporary settings if needed
    # if g.tempCommaMode:
    #     comma = True
    # else:
    #     comma = g.comma

    # output settings, which may be overrided by temp settings
    outputRadix = g.outputRadix
    integerGrouping = g.integerGrouping
    leadingZero = g.leadingZero

    if g.tempHexMode:
        integerGrouping = 4
        leadingZero = True
        outputRadix = 16

    if g.tempLeadingZeroMode:
        leadingZero = True

    if g.tempOctalMode:
        integerGrouping = 3
        leadingZero = True
        outputRadix = 8

    mpOutput = mpmathify( output )

    imaginary = im( mpOutput )
    real = re( mpOutput )

    result, negative = formatNumber( real, outputRadix, leadingZero, integerGrouping )

    if negative:
        result = '-' + result

    if imaginary != 0:
        strImaginary, negativeImaginary = formatNumber( imaginary, outputRadix, leadingZero )
        result = '( ' + result + ( ' - ' if negativeImaginary else ' + ' ) + strImaginary + 'j )'

    #print( 'formatOutput 2:', output )
    #print( )

    return result


#******************************************************************************
#
#  formatListOutput
#
#******************************************************************************

def formatListOutput( result, level=0, indent=0, file=sys.stdout ):
    '''
    This function formats a list for output, taking into account the list
    format level, as specified by the -s option.
    '''
    indentString = ' ' * indent

    first = True

    if level < g.listFormatLevel:
        useIndent = True
        print( indentString + '[', end='', file=file )
        levelIndent = ' ' * ( level + 1 ) * 4
    else:
        useIndent = False
        print( indentString + '[ ', end='', file=file )

    #print( 'result type', type( result ) )

    for item in result:
        newString = ''

        if isinstance( item, ( list, RPNGenerator ) ):
            if first:
                first = False

                if useIndent:
                    print( file=file )
                    print( indentString + levelIndent, end='', file=file )
            else:
                if useIndent:
                    print( ',', file=file )
                    print( indentString + levelIndent, end='', file=file )
                else:
                    print( ', ', end='', file=file )

            formatListOutput( item, level + 1, file=file )
            continue

        if isinstance( item, str ):
            newString = item
        elif isinstance( item, RPNDateTime ):
            newString = formatDateTime( item )
        elif isinstance( item, RPNMeasurement ):
            newString = formatOutput( nstr( item.value, min_fixed=-g.maximumFixed - 1 ) )
            newString += ' ' + formatUnits( item )
        else:
            newString = formatOutput( str( item ) )

        if first:
            first = False
        else:
            print( ', ', end='', file=file )

        if useIndent:
            print( file=file )
            print( indentString + levelIndent + newString, end='', flush=True, file=file )
        else:
            print( newString, end='', flush=True, file=file )

    if useIndent:
        print( file=file )
        print( ' ' * level * 4 + ']', end='', file=file )
    else:
        print( ' ]', end='', file=file )

    if level == 0:
        print( file=file )


#******************************************************************************
#
#  formatUnits
#
#  For outputting an RPNMeasurement object, this method formats the units part
#  of the object, and returns the formatted string.
#
#******************************************************************************

def formatUnits( measurement ):
    value = measurement.value

    unitName = measurement.getUnitName( )

    if unitName is not None:
        unitString = ''

        if mpf( -1.0 ) < value > mpf( 1.0 ) or value == 0:
            tempString = measurement.getPluralUnitName( )
        else:
            tempString = unitName

        tempString = tempString.replace( '_null_unit*', '' )
        tempString = tempString.replace( '*_null_unit', '' )

        tempString = tempString.replace( '_null_unit/', '1/' )
        tempString = tempString.replace( '/_null_unit', '' )

        for c in tempString:
            if c == '_':
                unitString += ' '
            else:
                unitString += c

        return unitString

    unitString = ''

    units = measurement.units

    # now that we've expanded the compound units, let's format...
    for unit in units:
        if unit == '_null_unit':
            continue

        exponent = units[ unit ]

        if exponent > 0:
            if unitString != '':
                unitString += ' '

            if value == 1:
                unitString += unit
            else:
                unitString += g.unitOperators[ unit ].plural

            if exponent > 1:
                unitString += '^' + str( int( exponent ) )

    negativeUnits = ''

    if unitString == '':
        for unit in units:
            if unit == '_null_unit':
                continue

            exponent = units[ unit ]

            if exponent < 0:
                if negativeUnits != '':
                    negativeUnits += ' '

                negativeUnits += unit
                negativeUnits += '^' + str( int( exponent ) )
    else:
        for unit in units:
            if unit == '_null_unit':
                continue

            exponent = units[ unit ]

            if exponent < 0:
                if negativeUnits == '':
                    negativeUnits = ' per '
                else:
                    negativeUnits += ' '

                negativeUnits += unit

                if exponent > 1:
                    negativeUnits += '^' + str( int( exponent ) )
                elif exponent < -1:
                    negativeUnits += '^' + str( int( -exponent ) )

    result = ''

    # filter out the underscores
    for c in unitString + negativeUnits:
        if c == '_':
            result += ' '
        else:
            result += c

    return result


#******************************************************************************
#
#  formatDateTime
#
#******************************************************************************

def formatDateTime( datetime ):
    if not isinstance( datetime, RPNDateTime ):
        raise ValueError( 'expected RPNDateTime' )

    if datetime.getDateOnly( ):
        return datetime.formatDate( )
    else:
        # if datetime.microsecond:
        #     return datetime.format( 'YYYY-MM-DD HH:mm:ss.SSSSSS' )
        # else:
        return datetime.format( )


#******************************************************************************
#
#  printParagraph
#
#******************************************************************************

def printParagraph( text, indent = 0 ):
    lines = textwrap.wrap( text, g.lineLength - ( indent + 1 ) )

    for line in lines:
        print( ' ' * indent + line )


#******************************************************************************
#
#  printOperatorHelp
#
#******************************************************************************

def printOperatorHelp( term, operatorInfo, operatorHelp, regularOperator = True):
    if regularOperator:
        if operatorInfo.argCount == 1:
            print( 'n ', end = '' )
        elif operatorInfo.argCount == 2:
            print( 'n k ', end = '' )
        elif operatorInfo.argCount == 3:
            print( 'a b c ', end = '' )
        elif operatorInfo.argCount == 4:
            print( 'a b c d ', end = '' )
        elif operatorInfo.argCount == 5:
            print( 'a b c d e ', end = '' )

    aliasList = [ key for key in g.aliases if term == g.aliases[ key ] ]

    print( term + ' - ' + operatorHelp[ 1 ] )

    print( )

    if len( aliasList ) > 1:
        printParagraph( 'aliases:  ' + ', '.join( aliasList ) )
    elif len( aliasList ) == 1:
        printParagraph( 'alias:  ' + aliasList[ 0 ] )

    print( 'category: ' + operatorHelp[ 0 ] )

    if operatorHelp[ 2 ] == '' or operatorHelp[ 2 ] == '\n':
        print( )
        print( 'No further help is available.' )
        print( )
    else:
        print( operatorHelp[ 2 ] )

    if regularOperator:
        if len( operatorHelp ) > 3:
            if operatorHelp[ 3 ] == '' or operatorHelp[ 3 ] == '\n':
                print( 'No examples are available.' )
            else:
                print( term + ' examples:' )
                print( operatorHelp[ 3 ] )
        else:
            print( 'No examples are available.' )

    if len( operatorHelp ) > 4 and len( operatorHelp[ 4 ] ) > 0:
        print( 'see also:  ', end='' )

        bFirst = True

        for name in sorted( operatorHelp[ 4 ] ):
            if bFirst:
                bFirst = False
            else:
                print( ', ', end='' )

            print( name, end='' )

        print( )


#******************************************************************************
#
#  printCategoryHelp
#
#******************************************************************************

def printCategoryHelp( category, operators, listOperators, modifiers, operatorHelp ):
    if category in basicUnitTypes:
        units = [ ]

        for unit, unitInfo in g.unitOperators.items( ):
            if unitInfo.unitType == category:
                units.append( unit )

        printParagraph( 'The ' + category + ' category includes the following units:' )
        print( )

        printParagraph( ', '.join( sorted( units ) ), indent=4 )
        return

    printParagraph( 'The ' + category + ' category includes the following operators (with aliases in parentheses):' )
    print( )

    operatorList = [ key for key in operators if operatorHelp[ key ][ 0 ] == category ]
    operatorList.extend( [ key for key in listOperators if operatorHelp[ key ][ 0 ] == category ] )
    operatorList.extend( [ key for key in modifiers if operatorHelp[ key ][ 0 ] == category ] )

    addAliases( operatorList, g.aliases )

    for operator in sorted( operatorList ):
        print( operator )


#******************************************************************************
#
#  printHelp
#
#******************************************************************************

def printHelp( terms = None, interactive = False ):
    from rpn.rpnOperators import constants, listOperators, modifiers, operators

    if terms is None:
        terms = [ ]

    loadHelpData( )
    loadUnitData( )

    if g.helpVersion != PROGRAM_VERSION:
        print( 'rpn:  help file version mismatch' )

    unitType = False

    if len( terms ) > 0:
        term = terms[ 0 ]
    else:
        term = ''

    if len( terms ) > 1 and terms[ 0 ] == 'unit_type':
        term = terms[ 1 ]
        unitType = True

    if term == '':
        if interactive:
            printInteractiveHelp( )
        else:
            printGeneralHelp( )
        return

    # first check if the term is an alias and translate
    if term in g.aliases:
        term = g.aliases[ term ]

    # then look for exact matches in all the lists of terms for which we have help support
    if term in operators and not unitType:
        printOperatorHelp( term, operators[ term ], g.operatorHelp[ term ] )
    elif term in g.unitOperators:
        printOperatorHelp( term, g.unitOperators[ term ], g.operatorHelp[ term ], regularOperator = False )
    elif term in g.constantOperators:
        printOperatorHelp( term, g.constantOperators[ term ], g.operatorHelp[ term ], regularOperator = False )
    elif term in constants:
        printOperatorHelp( term, constants[ term ], g.operatorHelp[ term ] )
    elif term in listOperators:
        printOperatorHelp( term, listOperators[ term ], g.operatorHelp[ term ] )
    elif term in modifiers:
        printOperatorHelp( term, modifiers[ term ], g.operatorHelp[ term ] )
    elif term in g.helpTopics:
        print( g.helpTopics[ term ] )
    elif term in g.operatorCategories:
        printCategoryHelp( term, operators, listOperators, modifiers, g.operatorHelp )
    elif term == 'unit_types':
        printParagraph( ', '.join( sorted( [ key for key in g.unitTypeDict.keys( ) if key != '_null_type' ] ) ), \
                        indent=4 )
    elif term in g.unitTypeDict:
        unitList = sorted( g.unitTypeDict[ term ] )
        addAliases( unitList, g.aliases )
        for unit in unitList:
            printParagraph( unit, indent=4 )
    else:
        # if no exact matches for any topic, let's look for partial matches
        if 'unit_types'.startswith( term ):
            print( 'Interpreting topic as \'unit_types\'.' )
            printParagraph( ', '.join( sorted( g.unitTypeDict.keys( ) ) ), indent=4 )
            return

        helpTerm = next( ( i for i in g.unitTypeDict if i != term and i.startswith( term ) ), '' )

        if helpTerm != '':
            print( )
            print( 'Interpreting topic as \'' + helpTerm + '\'.' )
            printParagraph( ', '.join( sorted( g.unitTypeDict[ helpTerm ] ) ), indent=4 )
            return

        helpTerm = next( ( i for i in operators if i != term and i.startswith( term ) ), '' )

        if helpTerm != '':
            print( 'Interpreting topic as \'' + helpTerm + '\'.' )
            print( )
            printOperatorHelp( helpTerm, operators[ helpTerm ], g.operatorHelp[ helpTerm ] )
            return

        helpTerm = next( ( i for i in constants if i != term and i.startswith( term ) ), '' )

        if helpTerm != '':
            print( 'Interpreting topic as \'' + helpTerm + '\'.' )
            print( )
            printOperatorHelp( helpTerm, constants[ helpTerm ], g.operatorHelp[ helpTerm ] )
            return

        helpTerm = next( ( i for i in listOperators if i != term and i.startswith( term ) ), '' )

        if helpTerm != '':
            print( 'Interpreting topic as \'' + helpTerm + '\'.' )
            print( )
            printOperatorHelp( helpTerm, listOperators[ helpTerm ], g.operatorHelp[ helpTerm ] )
            return

        helpTerm = next( ( i for i in modifiers if i != term and i.startswith( term ) ), '' )

        if helpTerm != '':
            print( 'Interpreting topic as \'' + helpTerm + '\'.' )
            print( )
            printOperatorHelp( helpTerm, modifiers[ helpTerm ], g.operatorHelp[ helpTerm ] )
            return

        helpTerm = next( ( i for i in g.helpTopics if i != term and i.startswith( term ) ), '' )

        if helpTerm != '':
            print( 'Interpreting topic as \'' + helpTerm + '\'.' )
            print( )
            print( g.helpTopics[ helpTerm ] )
            return

        helpTerm = next( ( i for i in g.operatorCategories if i != term and i.startswith( term ) ), '' )

        if helpTerm != '':
            print( 'Interpreting topic as \'' + helpTerm + '\'.' )
            print( )
            printCategoryHelp( helpTerm, operators, listOperators, modifiers, g.operatorHelp )
        else:
            print( 'Help topic not found.' )


#******************************************************************************
#
#  printGeneralHelp
#
#******************************************************************************

def printGeneralHelp( ):
    print( RPN_PROGRAM_NAME + ' - ' + PROGRAM_DESCRIPTION )
    print( COPYRIGHT_MESSAGE )
    print( )

    printParagraph(
        'For help on a specific topic, use \'rpn help\' and add a help topic, '
        'operator category or a specific operator name.' )

    printHelpTopics( )


#******************************************************************************
#
#  printInteractiveHelp
#
#******************************************************************************

def printInteractiveHelp( ):
    loadHelpData( )

    printParagraph(
        'For help on a specific topic, use the topic operator with a general topic, '
        'operator category or a specific operator name.' )

    printHelpTopics( )


#******************************************************************************
#
#  printHelpTopics
#
#******************************************************************************

def printHelpTopics( ):
    print( )
    print( 'The following is a list of general topics:' )
    print( )

    helpTopics = list( g.helpTopics.keys( ) )
    helpTopics.append( 'unit_types' )

    printParagraph( ', '.join( sorted( helpTopics ) ), indent=4 )

    print( )
    print( 'The following is a list of operator categories:' )
    print( )

    operatorCategories = set( g.operatorCategories )

    unitTypes = dict( basicUnitTypes )

    for i in unitTypes:
        if i in operatorCategories:
            operatorCategories.remove( i )

    printParagraph( ', '.join( sorted( operatorCategories ) ), indent=4 )

    print( )
    print( 'The following is a list of unit categories\n(use \'rpn help unit_type name\' for more info):' )
    print( )

    del unitTypes[ '_null_type' ]

    printParagraph( ', '.join( sorted( unitTypes ) ), indent=4 )


#******************************************************************************
#
#  printHelpModeHelp
#
#******************************************************************************

def printHelpModeHelp( ):
    printParagraph( 'rpn help mode - \'topics\' for a list of topics, \'exit\' to return to rpn' )


#******************************************************************************
#
#  printTitleScreen
#
#******************************************************************************

def printTitleScreen( programName, programDescription, showHelp=True ):
    print( programName + PROGRAM_VERSION_STRING + ' - ' + programDescription )
    print( COPYRIGHT_MESSAGE )

    if showHelp:
        print( )
        print( 'Type "help" for more information, and "exit" to exit.' )




#!/usr/bin/env python

#******************************************************************************
#
#  rpnTestUtils.py
#
#  rpnChilada test utility functions
#  copyright (c) 2020, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

import shlex
import time

from mpmath import almosteq, fsub, isinf, mpf, mpmathify, log10, mp, nan, workdps

from rpn.rpn import rpn, handleOutput
from rpn.rpnGenerator import RPNGenerator
from rpn.rpnMeasurement import RPNMeasurement

import rpn.rpnGlobals as g

if not hasattr( time, 'time_ns' ):
    from rpn.rpnNanoseconds import time_ns
else:
    from time import time_ns


#******************************************************************************
#
#  compareResults
#
#******************************************************************************

#import pysnooper
#@pysnooper.snoop( )
def compareResults( result1, result2 ):
    '''Compares two RPN expressions to make sure they produce the same result.
    Does nothing if the results compare successfully, otherwise raises an
    exception.'''

    if isinstance( result1, RPNGenerator ):
        return compareResults( list( result1.getGenerator( ) ), result2 )

    if isinstance( result2, RPNGenerator ):
        return compareResults( result1, list( result2.getGenerator( ) ) )

    if isinstance( result1, list ) != isinstance( result2, list ):
        print( '**** error in results comparison' )
        print( '    result 1: ', result1 )
        print( '    result 2: ', result2 )
        raise ValueError( 'one result is a list, the other isn\'t' )

    if isinstance( result1, str ) != isinstance( result2, str ):
        print( '**** error in results comparison' )
        print( '    result 1: ', result1 )
        print( '    result 2: ', result2 )
        raise ValueError( 'one result is a string, the other isn\'t' )

    if isinstance( result1, str ) and isinstance( result2, str ):
        if result1 != result2:
            print( '**** error in results comparison' )
            print( type( result1 ), type( result2 ) )
            print( result1, result2, 'are not equal' )
            raise ValueError( 'unit test failed' )

        return True

    if isinstance( result1, RPNMeasurement ) and isinstance( result2, RPNMeasurement ):
        #print( result1.value, result1.units )
        #print( result2.value, result2.units )

        if result1 != result2:
            print( '**** error in results comparison' )
            print( type( result1 ), type( result2 ) )

            with workdps( 50 ):
                print( result1.value, result1.units, result2.value, result2.units, 'are not equal' )

            raise ValueError( 'unit test failed' )

        return True

    if isinstance( result1, list ) and isinstance( result2, list ):
        if len( result1 ) != len( result2 ):
            raise ValueError( 'lists are not of equal length:', len( result1 ), len( result2 ) )

        for i in range( 0, len( result1 ) ):
            if isinf( result1[ i ] ):
                if isinf( result2[ i ] ):
                    return True
                else:
                    print( '**** error in results comparison' )
                    print( type( result1[ i ] ), type( result2[ i ] ) )
                    print( result1[ i ], result2[ i ], 'are not equal' )

                    raise ValueError( 'unit test failed' )

            if not compareValues( result1[ i ], result2[ i ] ):
                digits = max( log10( result1[ i ] ), log10( result2[ i ] ) ) + 5

                mp.dps = digits

                print( '**** error in results comparison' )
                print( type( result1[ i ] ), type( result2[ i ] ) )
                print( result1[ i ], result2[ i ], 'are not equal' )
                print( 'difference', fsub( result1[ i ], result2[ i ] ) )
                print( 'difference found at index', i )

                raise ValueError( 'unit test failed' )
    else:
        if not compareValues( result1, result2 ):
            print( '**** error in results comparison' )
            print( '    result 1: ', result1 )
            print( '    result 2: ', result2 )

            raise ValueError( 'unit test failed' )

    return True


#******************************************************************************
#
#  compareValues
#
#******************************************************************************

def compareValues( result1, result2 ):
    if isinstance( result1, RPNMeasurement ) != isinstance( result2, RPNMeasurement ):
        return False

    if isinstance( result1, RPNMeasurement ):
        return result1.__eq__( result2 )
    else:
        if isinf( result1 ):
            if isinf( result2 ):
                return True
            else:
                print( '**** error in results comparison' )
                print( type( result1 ), type( result2 ) )
                print( result1, result2, 'are not equal' )

                raise ValueError( 'unit test failed' )

        return almosteq( result1, result2 )


#******************************************************************************
#
#  expectException
#
#******************************************************************************

def expectException( command ):
    if g.testFilter:
        if g.testFilter not in command:
            return True

    print( 'rpn', command )

    result = rpn( shlex.split( command + ' -I' ) )

    if result == [ nan ]:
        print( 'exception test passed!' )
        print( '' )
        return True

    raise ValueError( 'exception was expected but didn\'t happen' )


#******************************************************************************
#
#  expectEqual
#
#******************************************************************************

#@pysnooper.snoop( )
def expectEqual( command1, command2 ):
    if g.testFilter:
        if g.testFilter not in command1 and g.testFilter not in command2:
            return

    if g.timeIndividualTests:
        startTime = time_ns( )

    print( 'rpn', command1 )
    print( 'rpn', command2 )

    # Converting to a list makes sure generators get evaluated before the
    # precision gets reset.
    result1 = rpn( shlex.split( command1 + ' -I' ) )[ 0 ]

    if isinstance( result1, RPNGenerator ):
        result1 = list( result1.getGenerator( ) )

    result2 = rpn( shlex.split( command2 + ' -I' ) )[ 0 ]

    if isinstance( result2, RPNGenerator ):
        result2 = list( result2.getGenerator( ) )

    compareResults( result1, result2 )

    print( '    both are equal!' )

    if g.timeIndividualTests:
        print( 'Test complete.  Time elapsed:  {:.3f} seconds'.format( ( time_ns( ) - startTime ) / 1000000000 ) )

    print( )


#******************************************************************************
#
#  areListsEquivalent
#
#******************************************************************************

def areListsEquivalent( list1, list2 ):
    if len( list1 ) != len( list2 ):
        raise ValueError( 'lists are not of equal length:', len( list1 ), len( list2 ) )

    temp2 = list( list2 )   # make a mutable copy

    try:
        for elem in list1:
            for elem2 in temp2:
                if compareValues( elem, elem2 ):
                    temp2.remove( elem2 )
                    break
    except ValueError:
        print( 'catch', elem, temp2 )
        return False

    return not temp2


#******************************************************************************
#
#  expectEquivalent
#
#  This is the same as expectEqual, but the results don't need to be in the
#  same order.
#
#******************************************************************************

def expectEquivalent( command1, command2 ):
    if g.testFilter:
        if g.testFilter not in command1 and g.testFilter not in command2:
            return

    if g.timeIndividualTests:
        startTime = time_ns( )

    print( 'rpn', command1 )
    print( 'rpn', command2 )

    result1 = rpn( shlex.split( command1 + ' -I' ) )[ 0 ]
    result2 = rpn( shlex.split( command2 + ' -I' ) )[ 0 ]

    if isinstance( result1, list ) != isinstance( result2, list ):
        print( '**** error in results comparison' )
        print( '    result 1: ', result1 )
        print( '    result 2: ', result2 )
        raise ValueError( 'one result is a list, the other isn\'t' )

    if not areListsEquivalent( result1, result2 ):
        print( '**** error in results comparison' )
        print( '    result 1: ', result1 )
        print( '    result 2: ', result2 )
        raise ValueError( 'unit test failed' )

    print( '    both are equal!' )

    if g.timeIndividualTests:
        print( 'Test complete.  Time elapsed:  {:.3f} seconds'.format( ( time_ns( ) - startTime ) / 1000000000 ) )

    print( '' )


#******************************************************************************
#
#  testOperator
#
#******************************************************************************

def testOperator( command, ignoreCache = True ):
    if g.testFilter:
        if g.testFilter not in command:
            return

    if g.timeIndividualTests:
        startTime = time_ns( )

    print( 'rpn', command )

    if ignoreCache:
        result = rpn( shlex.split( command + ' -I' ) )
    else:
        result = rpn( shlex.split( command ) )

    if result is not None and isinstance( result[ 0 ], mpf ) and result == [ nan ]:
        raise ValueError( 'unit test failed' )

    if result is not None:
        handleOutput( result )

    print( '    operator works!' )

    if g.timeIndividualTests:
        print( 'Test complete.  Time elapsed:  {:.3f} seconds'.format( ( time_ns( ) - startTime ) / 1000000000 ) )

    print( '' )


#******************************************************************************
#
#  expectResult
#
#******************************************************************************

def expectResult( command, expected ):
    if g.testFilter:
        if g.testFilter not in command:
            return

    if g.timeIndividualTests:
        startTime = time_ns( )

    print( 'rpn', command )
    result = rpn( shlex.split( command + ' -I' ) )[ 0 ]

    compare = None

    if isinstance( expected, list ):
        compare = [ ]

        for i in expected:
            if isinstance( i, ( int, float, complex ) ):
                compare.append( mpmathify( i ) )
            else:
                compare.append( i )
    else:
        if isinstance( expected, ( int, float, complex ) ):
            compare = mpmathify( expected )
        else:
            compare = expected

    compareResults( result, compare )

    print( '    test passed!' )

    if g.timeIndividualTests:
        print( 'Test complete.  Time elapsed:  {:.3f} seconds'.format( ( time_ns( ) - startTime ) / 1000000000 ) )

    print( '' )




#!/usr/bin/env python

#******************************************************************************
#
#  rpnComputer.py
#
#  rpnChilada computing operators
#  copyright (c) 2020, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

import struct

from mpmath import fadd, fdiv, floor, fmod, fmul, fsub, fsum, log, mpf, mp, mpmathify, power

from rpn.rpnGenerator import RPNGenerator
from rpn.rpnMeasurement import RPNMeasurement
from rpn.rpnSettings import setAccuracy
from rpn.rpnUtils import oneArgFunctionEvaluator, twoArgFunctionEvaluator, validateReal, validateRealInt

import rpn.rpnGlobals as g


#******************************************************************************
#
#  convertToSignedInt
#
#  two's compliment logic is in effect here
#
#******************************************************************************

def convertToSignedInt( n, k ):
    newPrecision = ( int( validateRealInt( k ) ) * 3 / 10 ) + 3

    if mp.dps < newPrecision:
        setAccuracy( newPrecision )

    value = fadd( validateRealInt( n ), ( power( 2, fsub( k, 1 ) ) ) )
    value = fmod( value, power( 2, k ) )
    value = fsub( value, ( power( 2, fsub( k, 1 ) ) ) )

    return value

@twoArgFunctionEvaluator( )
def convertToSignedIntOperator( n, k ):
    return convertToSignedInt( n, k )

@oneArgFunctionEvaluator( )
def convertToChar( n ):
    return convertToSignedInt( n, 8 )

@oneArgFunctionEvaluator( )
def convertToShort( n ):
    return convertToSignedInt( n, 16 )

@oneArgFunctionEvaluator( )
def convertToLong( n ):
    return convertToSignedInt( n, 32 )

@oneArgFunctionEvaluator( )
def convertToLongLong( n ):
    return convertToSignedInt( n, 64 )

@oneArgFunctionEvaluator( )
def convertToQuadLong( n ):
    return convertToSignedInt( n, 128 )


#******************************************************************************
#
#  getInvertedBits
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getInvertedBits( n ):
    value = validateRealInt( n )

    # determine how many groups of bits we will be looking at
    if value == 0:
        groupings = 1
    else:
        groupings = int( fadd( floor( fdiv( ( log( value, 2 ) ), g.bitwiseGroupSize ) ), 1 ) )

    placeValue = mpmathify( 1 << g.bitwiseGroupSize )
    multiplier = mpmathify( 1 )
    remaining = value

    result = mpmathify( 0 )

    for _ in range( 0, groupings ):
        # Let's let Python do the actual inverting
        group = fmod( ~int( fmod( remaining, placeValue ) ), placeValue )

        result += fmul( group, multiplier )

        remaining = floor( fdiv( remaining, placeValue ) )
        multiplier = fmul( multiplier, placeValue )

    return result


#******************************************************************************
#
#  performBitwiseOperation
#
#  The operations are performed on groups of bits as specified by the variable
#  bitwiseGroupSize.  Although doing it this way isn't really necessary, it
#  does mean that under normal circumstances the regular Python bit operators
#  can be used.
#
#******************************************************************************

def performBitwiseOperation( i, j, operation ):
    value1 = validateRealInt( i )
    value2 = validateRealInt( j )

    # determine how many groups of bits we will be looking at
    if value1 == 0:
        groupings = 1
    else:
        groupings = int( fadd( floor( fdiv( ( log( value1, 2 ) ), g.bitwiseGroupSize ) ), 1 ) )

    if value2 == 0:
        groupings2 = 1
    else:
        groupings2 = int( fadd( floor( fdiv( ( log( value2, 2 ) ), g.bitwiseGroupSize ) ), 1 ) )

    if groupings2 > groupings:
        groupings = groupings2

    placeValue = mpmathify( 1 << g.bitwiseGroupSize )
    multiplier = mpmathify( 1 )
    remaining1 = value1
    remaining2 = value2

    result = mpmathify( 0 )

    for _ in range( 0, groupings ):
        mod1 = fmod( remaining1, placeValue )
        mod2 = fmod( remaining2, placeValue )

        result = fadd( fmul( operation( int( mod1 ), int( mod2 ) ), multiplier ), result )

        remaining1 = floor( fdiv( remaining1, placeValue ) )
        remaining2 = floor( fdiv( remaining2, placeValue ) )

        multiplier = fmul( multiplier, placeValue )

    return result

@twoArgFunctionEvaluator( )
def getBitwiseAnd( n, k ):
    return performBitwiseOperation( n, k, lambda x, y: x & y )

@twoArgFunctionEvaluator( )
def getBitwiseNand( n, k ):
    return getInvertedBits( performBitwiseOperation( n, k, lambda x, y: x & y ) )

@twoArgFunctionEvaluator( )
def getBitwiseNor( n, k ):
    return getInvertedBits( performBitwiseOperation( n, k, lambda x, y: x | y ) )

@twoArgFunctionEvaluator( )
def getBitwiseOr( n, k ):
    return performBitwiseOperation( n, k, lambda x, y: x | y )

@twoArgFunctionEvaluator( )
def getBitwiseXor( n, k ):
    return performBitwiseOperation( n, k, lambda x, y: x ^ y )

@twoArgFunctionEvaluator( )
def getBitwiseXnor( n, k ):
    return getInvertedBit( performBitwiseOperation( n, k, lambda x, y: x ^ y ) )

@twoArgFunctionEvaluator( )
def shiftLeft( n, k ):
    return fmul( n, 1 << int( k ) )

@twoArgFunctionEvaluator( )
def shiftRight( n, k ):
    return floor( fdiv( n, 1 << int( k ) ) )


#******************************************************************************
#
#  getBitCount
#
#******************************************************************************

def getBitCount( n ):
    n = validateRealInt( n )

    if validateRealInt( n ) < 0:
        raise ValueError( '\'bit_count\' requires a positive integer value' )

    result = 0

    if isinstance( n, RPNMeasurement ):
        value = n.value
    else:
        value = n

    while value:
        result += fmod( value, 2 )
        value = floor( fdiv( value, 2 ) )

    return result

@oneArgFunctionEvaluator( )
def getBitCountOperator( n ):
    return getBitCount( n )

@oneArgFunctionEvaluator( )
def getParity( n ):
    return fmod( getBitCount( n ), 2 )


#******************************************************************************
#
#  unpackInteger
#
#******************************************************************************

def unpackInteger( n, fields ):
    if isinstance( n, RPNGenerator ):
        return unpackInteger( list( n ), fields )
    elif isinstance( n, list ):
        return [ unpackInteger( i, fields ) for i in n ]

    if isinstance( fields, RPNGenerator ):
        return unpackInteger( n, list( fields ) )
    elif not isinstance( fields, list ):
        return unpackInteger( n, [ fields ] )

    value = validateRealInt( n )
    result = [ ]

    for field in reversed( fields ):
        size = int( field )
        result.insert( 0, value & ( 2 ** size - 1 ) )
        value >>= size

    return result


#******************************************************************************
#
#  packInteger
#
#******************************************************************************

def packInteger( values, fields ):
    if isinstance( values, RPNGenerator ):
        return packInteger( list( values ), fields )
    elif not isinstance( values, list ):
        return unpackInteger( [ values ], fields )

    if isinstance( fields, RPNGenerator ):
        return packInteger( values, list( fields ) )
    elif not isinstance( fields, list ):
        return unpackInteger( values, [ fields ] )

    if isinstance( values[ 0 ], list ):
        return [ unpackInteger( value, fields ) for value in values ]

    result = 0

    count = min( len( values ), len( fields ) )

    size = 0

    for i in range( count, 0, -1 ):
        field = int( fields[ i - 1 ] )
        value = int( values[ i - 1 ] ) & ( 2 ** field - 1 )
        result = fadd( result, fmul( value, power( 2, size ) ) )
        size += field

    return result


#******************************************************************************
#
#  interpretAsFloat
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def interpretAsFloat( n ):
    if ( validateRealInt( n ) < 0 ) or ( n >= 2 ** 32 - 1 ):
        raise ValueError( 'value out of range for a 32-bit float' )

    intValue = struct.pack( 'I', int( n ) )
    return mpf( struct.unpack( 'f', intValue )[ 0 ] )


#******************************************************************************
#
#  interpretAsDouble
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def interpretAsDouble( n ):
    setAccuracy( 25 )

    if ( validateRealInt( n ) < 0 ) or ( n > 2 ** 64 - 1 ):
        raise ValueError( 'value out of range for a 64-bit float' )

    intValue = struct.pack( 'Q', int( n ) )
    return mpf( struct.unpack( 'd', intValue )[ 0 ] )


@oneArgFunctionEvaluator( )
def convertToDouble( n ):
    return fsum( b << 8 * i for i, b in enumerate( struct.pack( 'd', float( validateReal( n ) ) ) ) )

@oneArgFunctionEvaluator( )
def convertToFloat( n ):
    return fsum( b << 8 * i for i, b in enumerate( struct.pack( 'f', float( validateReal( n ) ) ) ) )

@twoArgFunctionEvaluator( )
def convertToUnsignedInt( n, k ):
    return fmod( validateRealInt( n ), power( 2, validateReal( k ) ) )

@oneArgFunctionEvaluator( )
def convertToUnsignedChar( n ):
    return fmod( validateRealInt( n ), power( 2, 8 ) )

@oneArgFunctionEvaluator( )
def convertToUnsignedShort( n ):
    return fmod( validateRealInt( n ), power( 2, 16 ) )

@oneArgFunctionEvaluator( )
def convertToUnsignedLong( n ):
    return fmod( validateRealInt( n ), power( 2, 32 ) )

@oneArgFunctionEvaluator( )
def convertToUnsignedLongLong( n ):
    return fmod( validateRealInt( n ), power( 2, 64 ) )

@oneArgFunctionEvaluator( )
def convertToUnsignedQuadLong( n ):
    return fmod( validateRealInt( n ), power( 2, 128 ) )

@twoArgFunctionEvaluator( )
def andOperands( n, k ):
    return 1 if ( n != 0 and k != 0 ) else 0

@oneArgFunctionEvaluator( )
def notOperand( n ):
    return 1 if n == 0 else 0

@twoArgFunctionEvaluator( )
def nandOperands( n, k ):
    return 0 if ( n != 0 and k != 0 ) else 1

@twoArgFunctionEvaluator( )
def norOperands( n, k ):
    return 0 if ( n != 0 or k != 0 ) else 1

@twoArgFunctionEvaluator( )
def orOperands( n, k ):
    return 1 if ( n != 0 or k != 0 ) else 0

@twoArgFunctionEvaluator( )
def xnorOperands( n, k ):
    return 1 if ( n != 0 ) == ( k != 0 ) else 0

@twoArgFunctionEvaluator( )
def xorOperands( n, k ):
    return 1 if ( n != 0 ) != ( k != 0 ) else 0

